@@ -8,7 +8,7 @@
    This file is part of Valgrind, a dynamic binary instrumentation
    framework.
 
-   Copyright (C) 2000-2007 Julian Seward 
+   Copyright (C) 2000-2013 Julian Seward 
       jseward@acm.org
 
    This program is free software; you can redistribute it and/or
@@ -32,6 +32,16 @@
 
 /*---------------------------------------------------------------*/
 
+/* Include valgrind headers before system headers to avoid problems
+   with the system headers #defining things which are used as names
+   of structure members in vki headers. */
+
+#include "pub_core_basics.h"
+#include "pub_core_libcassert.h"    // For VG_BUGS_TO
+#include "pub_core_vki.h"           // Avoids warnings from 
+                                    // pub_core_libcfile.h
+#include "pub_core_libcfile.h"      // For VG_CLO_DEFAULT_LOGPORT
+
 #include <stdio.h>
 #include <unistd.h>
 #include <string.h>
@@ -44,12 +54,6 @@
 #include <sys/socket.h>
 #include <netinet/in.h>
 
-#include "pub_core_basics.h"
-#include "pub_core_libcassert.h"    // For VG_BUGS_TO
-#include "pub_core_vki.h"           // Avoids warnings from 
-                                    // pub_core_libcfile.h
-#include "pub_core_libcfile.h"      // For VG_CLO_DEFAULT_LOGPORT
-
 
 /*---------------------------------------------------------------*/
 
@@ -60,7 +64,7 @@
 /*---------------------------------------------------------------*/
 
 __attribute__ ((noreturn))
-static void panic ( Char* str )
+static void panic ( const char* str )
 {
    fprintf(stderr,
            "\nvalgrind-listener: the "
@@ -71,7 +75,7 @@ static void panic ( Char* str )
 }
 
 __attribute__ ((noreturn))
-static void my_assert_fail ( const Char* expr, const Char* file, Int line, const Char* fn )
+static void my_assert_fail ( const char* expr, const char* file, int line, const char* fn )
 {
    fprintf(stderr,
            "\nvalgrind-listener: %s:%d (%s): Assertion '%s' failed.\n",
@@ -128,7 +132,8 @@ static void copyout ( char* buf, int nbuf )
       if (buf[i] == '\n') {
          fprintf(stdout, "\n(%d) ", conn_count);
       } else {
-         fwrite(&buf[i], 1, 1, stdout);
+         __attribute__((unused)) size_t ignored 
+            = fwrite(&buf[i], 1, 1, stdout);
       }
    }
    fflush(stdout);
@@ -163,7 +168,7 @@ static void snooze ( void )
 
 
 /* returns 0 if invalid, else port # */
-static int atoi_portno ( char* str )
+static int atoi_portno ( const char* str )
 {
    int n = 0;
    while (1) {
@@ -205,7 +210,7 @@ static void usage ( void )
 }
 
 
-static void banner ( char* str )
+static void banner ( const char* str )
 {
    time_t t;
    t = time(NULL);
