@@ -6,7 +6,7 @@
 /*
    This file is part of Callgrind, a Valgrind tool for call tracing.
 
-   Copyright (C) 2002-2007, Josef Weidendorfer (Josef.Weidendorfer@gmx.de)
+   Copyright (C) 2002-2013, Josef Weidendorfer (Josef.Weidendorfer@gmx.de)
 
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
@@ -29,19 +29,19 @@
 #include "config.h"
 #include "global.h"
 
-#include <pub_tool_threadstate.h>
-#include <pub_tool_libcfile.h>
+#include "pub_tool_threadstate.h"
+#include "pub_tool_libcfile.h"
 
 
 /* Dump Part Counter */
 static Int out_counter = 0;
 
-static Char* out_file = 0;
-static Char* out_directory = 0;
+static HChar* out_file = 0;
+static HChar* out_directory = 0;
 static Bool dumps_initialized = False;
 
 /* Command */
-static Char cmdbuf[BUF_LEN];
+static HChar cmdbuf[BUF_LEN];
 
 /* Total reads/writes/misses sum over all dumps and threads.
  * Updated during CC traversal at dump time.
@@ -55,22 +55,22 @@ EventMapping* CLG_(dumpmap) = 0;
  *  print_fn_pos, fprint_apos, fprint_fcost, fprint_jcc,
  *  fprint_fcc_ln, dump_run_info, dump_state_info
  */
-static Char outbuf[FILENAME_LEN + FN_NAME_LEN + OBJ_NAME_LEN];
+static HChar outbuf[FILENAME_LEN + FN_NAME_LEN + OBJ_NAME_LEN + COSTS_LEN];
 
 Int CLG_(get_dump_counter)(void)
 {
   return out_counter;
 }
 
-Char* CLG_(get_out_file)()
+HChar* CLG_(get_out_file)()
 {
-    CLG_ASSERT(dumps_initialized);
+    CLG_(init_dumps)();
     return out_file;
 }
 
-Char* CLG_(get_out_directory)()
+HChar* CLG_(get_out_directory)()
 {
-    CLG_ASSERT(dumps_initialized);
+    CLG_(init_dumps)();
     return out_directory;
 }
 
@@ -105,7 +105,8 @@ void init_dump_array(void)
       CLG_(stat).distinct_fns +
       CLG_(stat).context_counter;
     CLG_ASSERT(dump_array == 0);
-    dump_array = (Bool*) CLG_MALLOC(dump_array_size * sizeof(Bool));
+    dump_array = (Bool*) CLG_MALLOC("cl.dump.ida.1",
+                                    dump_array_size * sizeof(Bool));
     obj_dumped  = dump_array;
     file_dumped = obj_dumped + CLG_(stat).distinct_objs;
     fn_dumped   = file_dumped + CLG_(stat).distinct_files;
@@ -144,15 +145,15 @@ void init_fpos(FnPos* p)
 
 #if 0
 static __inline__
-static void my_fwrite(Int fd, Char* buf, Int len)
+static void my_fwrite(Int fd, const HChar* buf, Int len)
 {
-	VG_(write)(fd, (void*)buf, len);
+	VG_(write)(fd, buf, len);
 }
 #else
 
 #define FWRITE_BUFSIZE 32000
 #define FWRITE_THROUGH 10000
-static Char fwrite_buf[FWRITE_BUFSIZE];
+static HChar fwrite_buf[FWRITE_BUFSIZE];
 static Int fwrite_pos;
 static Int fwrite_fd = -1;
 
@@ -160,11 +161,11 @@ static __inline__
 void fwrite_flush(void)
 {
     if ((fwrite_fd>=0) && (fwrite_pos>0))
-	VG_(write)(fwrite_fd, (void*)fwrite_buf, fwrite_pos);
+	VG_(write)(fwrite_fd, fwrite_buf, fwrite_pos);
     fwrite_pos = 0;
 }
 
-static void my_fwrite(Int fd, Char* buf, Int len)
+static void my_fwrite(Int fd, const HChar* buf, Int len)
 {
     if (fwrite_fd != fd) {
 	fwrite_flush();
@@ -172,7 +173,7 @@ static void my_fwrite(Int fd, Char* buf, Int len)
     }
     if (len > FWRITE_THROUGH) {
 	fwrite_flush();
-	VG_(write)(fd, (void*)buf, len);
+	VG_(write)(fd, buf, len);
 	return;
     }
     if (FWRITE_BUFSIZE - fwrite_pos <= len) fwrite_flush();
@@ -182,21 +183,21 @@ static void my_fwrite(Int fd, Char* buf, Int len)
 #endif
 
 
-static void print_obj(Char* buf, obj_node* obj)
+static void print_obj(HChar* buf, obj_node* obj)
 {
-    int n;
+    //int n;
 
     if (CLG_(clo).compress_strings) {
 	CLG_ASSERT(obj_dumped != 0);
 	if (obj_dumped[obj->number])
-	    n = VG_(sprintf)(buf, "(%d)\n", obj->number);
+	    /*n =*/ VG_(sprintf)(buf, "(%d)\n", obj->number);
 	else {
-	    n = VG_(sprintf)(buf, "(%d) %s\n",
+	    /*n =*/ VG_(sprintf)(buf, "(%d) %s\n",
 			     obj->number, obj->name);
 	}
     }
     else
-	n = VG_(sprintf)(buf, "%s\n", obj->name);
+	/*n =*/ VG_(sprintf)(buf, "%s\n", obj->name);
 
 #if 0
     /* add mapping parameters the first time a object is dumped
@@ -211,7 +212,7 @@ static void print_obj(Char* buf, obj_node* obj)
 #endif
 }
 
-static void print_file(Char* buf, file_node* file)
+static void print_file(HChar* buf, file_node* file)
 {
     if (CLG_(clo).compress_strings) {
 	CLG_ASSERT(file_dumped != 0);
@@ -230,7 +231,7 @@ static void print_file(Char* buf, file_node* file)
 /*
  * tag can be "fn", "cfn", "jfn"
  */
-static void print_fn(Int fd, Char* buf, Char* tag, fn_node* fn)
+static void print_fn(Int fd, HChar* buf, const HChar* tag, fn_node* fn)
 {
     int p;
     p = VG_(sprintf)(buf, "%s=",tag);
@@ -250,7 +251,7 @@ static void print_fn(Int fd, Char* buf, Char* tag, fn_node* fn)
     my_fwrite(fd, buf, p);
 }
 
-static void print_mangled_fn(Int fd, Char* buf, Char* tag, 
+static void print_mangled_fn(Int fd, HChar* buf, const HChar* tag, 
 			     Context* cxt, int rec_index)
 {
     int p, i;
@@ -335,6 +336,8 @@ static Bool print_fn_pos(int fd, FnPos* last, BBCC* bbcc)
 {
     Bool res = False;
 
+    CLG_ASSERT(bbcc && bbcc->cxt);
+
     CLG_DEBUGIF(3) {
 	CLG_DEBUG(2, "+ print_fn_pos: ");
 	CLG_(print_cxt)(16, bbcc->cxt, bbcc->rec_index);
@@ -343,22 +346,22 @@ static Bool print_fn_pos(int fd, FnPos* last, BBCC* bbcc)
     if (!CLG_(clo).mangle_names) {
 	if (last->rec_index != bbcc->rec_index) {
 	    VG_(sprintf)(outbuf, "rec=%d\n\n", bbcc->rec_index);
-	    my_fwrite(fd, (void*)outbuf, VG_(strlen)(outbuf));
+	    my_fwrite(fd, outbuf, VG_(strlen)(outbuf));
 	    last->rec_index = bbcc->rec_index;
 	    last->cxt = 0; /* reprint context */
 	    res = True;
 	}
 	
 	if (last->cxt != bbcc->cxt) {
-	    fn_node* last_from = (last->cxt && last->cxt->size>1) ?
+	    fn_node* last_from = (last->cxt && last->cxt->size >1) ?
 				 last->cxt->fn[1] : 0;
-	    fn_node* curr_from = (bbcc->cxt && bbcc->cxt->size>1) ?
+	    fn_node* curr_from = (bbcc->cxt->size >1) ?
 				 bbcc->cxt->fn[1] : 0;
 	    if (curr_from == 0) {
 		if (last_from != 0) {
 		    /* switch back to no context */
 		    VG_(sprintf)(outbuf, "frfn=(spontaneous)\n");
-		    my_fwrite(fd, (void*)outbuf, VG_(strlen)(outbuf));
+		    my_fwrite(fd, outbuf, VG_(strlen)(outbuf));
 		    res = True;
 		}
 	    }
@@ -373,7 +376,7 @@ static Bool print_fn_pos(int fd, FnPos* last, BBCC* bbcc)
     if (last->obj != bbcc->cxt->fn[0]->file->obj) {
 	VG_(sprintf)(outbuf, "ob=");
 	print_obj(outbuf+3, bbcc->cxt->fn[0]->file->obj);
-	my_fwrite(fd, (void*)outbuf, VG_(strlen)(outbuf));
+	my_fwrite(fd, outbuf, VG_(strlen)(outbuf));
 	last->obj = bbcc->cxt->fn[0]->file->obj;
 	res = True;
     }
@@ -381,7 +384,7 @@ static Bool print_fn_pos(int fd, FnPos* last, BBCC* bbcc)
     if (last->file != bbcc->cxt->fn[0]->file) {
 	VG_(sprintf)(outbuf, "fl=");
 	print_file(outbuf+3, bbcc->cxt->fn[0]->file);
-	my_fwrite(fd, (void*)outbuf, VG_(strlen)(outbuf));
+	my_fwrite(fd, outbuf, VG_(strlen)(outbuf));
 	last->file = bbcc->cxt->fn[0]->file;
 	res = True;
     }
@@ -437,8 +440,8 @@ void init_debug_cache(void)
 static /* __inline__ */
 Bool get_debug_pos(BBCC* bbcc, Addr addr, AddrPos* p)
 {
-    Char file[FILENAME_LEN];
-    Char dir[FILENAME_LEN];
+    HChar file[FILENAME_LEN];
+    HChar dir[FILENAME_LEN];
     Bool found_file_line, found_dirname;
 
     int cachepos = addr % DEBUG_CACHE_SIZE;
@@ -477,7 +480,7 @@ Bool get_debug_pos(BBCC* bbcc, Addr addr, AddrPos* p)
     p->addr = addr - bbcc->bb->obj->offset;
     p->bb_addr = bbcc->bb->offset;
 
-    CLG_DEBUG(3, "  get_debug_pos(%p): BB %p, fn '%s', file '%s', line %u\n",
+    CLG_DEBUG(3, "  get_debug_pos(%#lx): BB %#lx, fn '%s', file '%s', line %u\n",
 	     addr, bb_addr(bbcc->bb), bbcc->cxt->fn[0]->name,
 	     p->file->name, p->line);
 
@@ -519,7 +522,7 @@ static void init_fcost(AddrCost* c, Addr addr, Addr bbaddr, file_node* file)
 static void fprint_apos(Int fd, AddrPos* curr, AddrPos* last, file_node* func_file)
 {
     CLG_ASSERT(curr->file != 0);
-    CLG_DEBUG(2, "    print_apos(file '%s', line %d, bb %p, addr %p) fnFile '%s'\n",
+    CLG_DEBUG(2, "    print_apos(file '%s', line %d, bb %#lx, addr %#lx) fnFile '%s'\n",
 	     curr->file->name, curr->line, curr->bb_addr, curr->addr,
 	     func_file->name);
 
@@ -531,13 +534,13 @@ static void fprint_apos(Int fd, AddrPos* curr, AddrPos* last, file_node* func_fi
 	else
 	    VG_(sprintf)(outbuf, "fi=");
 	print_file(outbuf+3, curr->file);
-	my_fwrite(fd, (void*)outbuf, VG_(strlen)(outbuf));
+	my_fwrite(fd, outbuf, VG_(strlen)(outbuf));
     }
 
     if (CLG_(clo).dump_bbs) {
 	if (curr->line != last->line) {
 	    VG_(sprintf)(outbuf, "ln=%d\n", curr->line);
-	    my_fwrite(fd, (void*)outbuf, VG_(strlen)(outbuf));
+	    my_fwrite(fd, outbuf, VG_(strlen)(outbuf));
 	}
     }
 }
@@ -569,7 +572,7 @@ void fprint_pos(Int fd, AddrPos* curr, AddrPos* last)
 		    p = VG_(sprintf)(outbuf, "%d ", diff);
 	    }
 	    else
-		p = VG_(sprintf)(outbuf, "%p ", curr->addr);
+		p = VG_(sprintf)(outbuf, "%#lx ", curr->addr);
 	}
 
 	if (CLG_(clo).dump_bb) {
@@ -584,7 +587,7 @@ void fprint_pos(Int fd, AddrPos* curr, AddrPos* last)
 		    p += VG_(sprintf)(outbuf+p, "%d ", diff);
 	    }
 	    else
-		p += VG_(sprintf)(outbuf+p, "%p ", curr->bb_addr);
+		p += VG_(sprintf)(outbuf+p, "%#lx ", curr->bb_addr);
 	}
 
 	if (CLG_(clo).dump_line) {
@@ -603,7 +606,7 @@ void fprint_pos(Int fd, AddrPos* curr, AddrPos* last)
 		VG_(sprintf)(outbuf+p, "%u ", curr->line);
 	}
     }
-    my_fwrite(fd, (void*)outbuf, VG_(strlen)(outbuf));
+    my_fwrite(fd, outbuf, VG_(strlen)(outbuf));
 }
 
 
@@ -616,7 +619,7 @@ void fprint_cost(int fd, EventMapping* es, ULong* cost)
 {
   int p = CLG_(sprint_mappingcost)(outbuf, es, cost);
   VG_(sprintf)(outbuf+p, "\n");
-  my_fwrite(fd, (void*)outbuf, VG_(strlen)(outbuf));
+  my_fwrite(fd, outbuf, VG_(strlen)(outbuf));
   return;
 }
 
@@ -630,7 +633,7 @@ void fprint_cost(int fd, EventMapping* es, ULong* cost)
 static void fprint_fcost(Int fd, AddrCost* c, AddrPos* last)
 {
   CLG_DEBUGIF(3) {
-    CLG_DEBUG(2, "   print_fcost(file '%s', line %d, bb %p, addr %p):\n",
+    CLG_DEBUG(2, "   print_fcost(file '%s', line %d, bb %#lx, addr %#lx):\n",
 	     c->p.file->name, c->p.line, c->p.bb_addr, c->p.addr);
     CLG_(print_cost)(-5, CLG_(sets).full, c->cost);
   }
@@ -658,13 +661,15 @@ static void fprint_jcc(Int fd, jCC* jcc, AddrPos* curr, AddrPos* last, ULong eco
       CLG_(print_jcc)(-10, jcc);
     }
 
+    CLG_ASSERT(jcc->to !=0);
+    CLG_ASSERT(jcc->from !=0);
+    
     if (!get_debug_pos(jcc->to, bb_addr(jcc->to->bb), &target)) {
 	/* if we don't have debug info, don't switch to file "???" */
 	target.file = last->file;
     }
 
-    if (jcc->from &&
-	(jcc->jmpkind == JmpCond || jcc->jmpkind == Ijk_Boring)) {
+    if ((jcc->jmpkind == jk_CondJump) || (jcc->jmpkind == jk_Jump)) {
 	    
       /* this is a JCC for a followed conditional or boring jump. */
       CLG_ASSERT(CLG_(is_zero_cost)( CLG_(sets).full, jcc->cost));
@@ -687,7 +692,7 @@ static void fprint_jcc(Int fd, jCC* jcc, AddrPos* curr, AddrPos* last, ULong eco
 	if (last->file != target.file) {
 	    VG_(sprintf)(outbuf, "jfi=");
 	    print_file(outbuf+4, target.file);
-	    my_fwrite(fd, (void*)outbuf, VG_(strlen)(outbuf));
+	    my_fwrite(fd, outbuf, VG_(strlen)(outbuf));
 	}
 	
 	if (jcc->from->cxt != jcc->to->cxt) {
@@ -698,7 +703,7 @@ static void fprint_jcc(Int fd, jCC* jcc, AddrPos* curr, AddrPos* last, ULong eco
 		print_fn(fd, outbuf, "jfn", jcc->to->cxt->fn[0]);
 	}
 	    
-	if (jcc->jmpkind == JmpCond) {
+	if (jcc->jmpkind == jk_CondJump) {
 	    /* format: jcnd=<followed>/<executions> <target> */
 	    VG_(sprintf)(outbuf, "jcnd=%llu/%llu ",
 			 jcc->call_counter, ecounter);
@@ -708,7 +713,7 @@ static void fprint_jcc(Int fd, jCC* jcc, AddrPos* curr, AddrPos* last, ULong eco
 	    VG_(sprintf)(outbuf, "jump=%llu ",
 			 jcc->call_counter);
 	}
-	my_fwrite(fd, (void*)outbuf, VG_(strlen)(outbuf));
+	my_fwrite(fd, outbuf, VG_(strlen)(outbuf));
 		
 	fprint_pos(fd, &target, last);
 	my_fwrite(fd, "\n", 1);
@@ -719,8 +724,6 @@ static void fprint_jcc(Int fd, jCC* jcc, AddrPos* curr, AddrPos* last, ULong eco
 	return;
     }
 
-    CLG_ASSERT(jcc->to !=0);
-    
     file = jcc->to->cxt->fn[0]->file;
     obj  = jcc->to->bb->obj;
     
@@ -728,14 +731,14 @@ static void fprint_jcc(Int fd, jCC* jcc, AddrPos* curr, AddrPos* last, ULong eco
     if (jcc->from->cxt->fn[0]->file->obj != obj) {
 	VG_(sprintf)(outbuf, "cob=");
 	print_obj(outbuf+4, obj);
-	my_fwrite(fd, (void*)outbuf, VG_(strlen)(outbuf));
+	my_fwrite(fd, outbuf, VG_(strlen)(outbuf));
     }
 
     /* file of called position different to current file? */
     if (last->file != file) {
 	VG_(sprintf)(outbuf, "cfi=");
 	print_file(outbuf+4, file);
-	my_fwrite(fd, (void*)outbuf, VG_(strlen)(outbuf));
+	my_fwrite(fd, outbuf, VG_(strlen)(outbuf));
     }
 
     if (CLG_(clo).mangle_names)
@@ -746,7 +749,7 @@ static void fprint_jcc(Int fd, jCC* jcc, AddrPos* curr, AddrPos* last, ULong eco
     if (!CLG_(is_zero_cost)( CLG_(sets).full, jcc->cost)) {
       VG_(sprintf)(outbuf, "calls=%llu ", 
 		   jcc->call_counter);
-	my_fwrite(fd, (void*)outbuf, VG_(strlen)(outbuf));
+	my_fwrite(fd, outbuf, VG_(strlen)(outbuf));
 
 	fprint_pos(fd, &target, last);
 	my_fwrite(fd, "\n", 1);	
@@ -786,7 +789,7 @@ static Bool fprint_bbcc(Int fd, BBCC* bbcc, AddrPos* last)
   CLG_ASSERT(bbcc->cxt != 0);
   CLG_DEBUGIF(1) {
     VG_(printf)("+ fprint_bbcc (Instr %d): ", bb->instr_count);
-    CLG_(print_bbcc)(15, bbcc, False);
+    CLG_(print_bbcc)(15, bbcc);
   }
 
   CLG_ASSERT(currSum == 0 || currSum == 1);
@@ -831,7 +834,8 @@ static Bool fprint_bbcc(Int fd, BBCC* bbcc, AddrPos* last)
     if (bb->jmp[jmp].instr == instr) {
 	jcc_count=0;
 	for(jcc=bbcc->jmp[jmp].jcc_list; jcc; jcc=jcc->next_from)
-	    if ((jcc->jmpkind != Ijk_Call) && (jcc->call_counter >0))
+	    if (((jcc->jmpkind != jk_Call) && (jcc->call_counter >0)) ||
+		(!CLG_(is_zero_cost)( CLG_(sets).full, jcc->cost )))
 	      jcc_count++;
 
 	if (jcc_count>0) {    
@@ -844,7 +848,8 @@ static Bool fprint_bbcc(Int fd, BBCC* bbcc, AddrPos* last)
 	    fprint_apos(fd, &(currCost->p), last, bbcc->cxt->fn[0]->file);
 	    something_written = True;
 	    for(jcc=bbcc->jmp[jmp].jcc_list; jcc; jcc=jcc->next_from) {
-		if ((jcc->jmpkind != Ijk_Call) && (jcc->call_counter >0))
+		if (((jcc->jmpkind != jk_Call) && (jcc->call_counter >0)) ||
+		    (!CLG_(is_zero_cost)( CLG_(sets).full, jcc->cost )))
 		    fprint_jcc(fd, jcc, &(currCost->p), last, ecounter);
 	    }
 	}
@@ -862,7 +867,7 @@ static Bool fprint_bbcc(Int fd, BBCC* bbcc, AddrPos* last)
   jcc_count = 0;
   for(jcc=bbcc->jmp[jmp].jcc_list; jcc; jcc=jcc->next_from) {
       /* yes, if JCC only counts jmp arcs or cost >0 */
-      if ( ((jcc->jmpkind != Ijk_Call) && (jcc->call_counter >0)) ||
+      if ( ((jcc->jmpkind != jk_Call) && (jcc->call_counter >0)) ||
 	   (!CLG_(is_zero_cost)( CLG_(sets).full, jcc->cost )))
 	  jcc_count++;
   }
@@ -888,7 +893,7 @@ static Bool fprint_bbcc(Int fd, BBCC* bbcc, AddrPos* last)
 			      currCost->cost, bbcc->skipped );
 #if 0
       VG_(sprintf)(outbuf, "# Skipped\n");
-      my_fwrite(fd, (void*)outbuf, VG_(strlen)(outbuf));
+      my_fwrite(fd, outbuf, VG_(strlen)(outbuf));
 #endif
       fprint_fcost(fd, currCost, last);
     }
@@ -896,7 +901,7 @@ static Bool fprint_bbcc(Int fd, BBCC* bbcc, AddrPos* last)
     if (jcc_count > 0)
 	for(jcc=bbcc->jmp[jmp].jcc_list; jcc; jcc=jcc->next_from) {
 	    CLG_ASSERT(jcc->jmp == jmp);
-	    if ( ((jcc->jmpkind != Ijk_Call) && (jcc->call_counter >0)) ||
+	    if ( ((jcc->jmpkind != jk_Call) && (jcc->call_counter >0)) ||
 		 (!CLG_(is_zero_cost)( CLG_(sets).full, jcc->cost )))
 	  
 		fprint_jcc(fd, jcc, &(currCost->p), last, ecounter);
@@ -910,7 +915,7 @@ static Bool fprint_bbcc(Int fd, BBCC* bbcc, AddrPos* last)
       fprint_apos(fd, &(currCost->p), last, bbcc->cxt->fn[0]->file);
       fprint_fcost(fd, currCost, last);
     }
-    if (CLG_(clo).dump_bbs) my_fwrite(fd, (void*)"\n", 1);
+    if (CLG_(clo).dump_bbs) my_fwrite(fd, "\n", 1);
     
     /* when every cost was immediatly written, we must have done so,
      * as this function is only called when there's cost in a BBCC
@@ -1018,7 +1023,7 @@ static void qsort(BBCC **a, int n, int (*cmp)(BBCC**,BBCC**))
 	int s, r;
 	BBCC* v;
 
-	CLG_DEBUG(8, "  qsort(%d,%d)\n", a-qsort_start, n);
+	CLG_DEBUG(8, "  qsort(%ld,%ld)\n", a-qsort_start + 0L, n + 0L);
 
 	if (n < 7) {	 /* Insertion sort on smallest arrays */
 		for (pm = a+1; pm < a+n; pm++)
@@ -1027,7 +1032,8 @@ static void qsort(BBCC **a, int n, int (*cmp)(BBCC**,BBCC**))
 
 		CLG_DEBUGIF(8) {
 		    for (pm = a; pm < a+n; pm++) {
-			VG_(printf)("   %3d BB %p, ", pm - qsort_start,
+			VG_(printf)("   %3ld BB %#lx, ",
+                                    pm - qsort_start + 0L,
 				    bb_addr((*pm)->bb));      
 			CLG_(print_cxt)(9, (*pm)->cxt, (*pm)->rec_index);
 		    }
@@ -1081,25 +1087,30 @@ static void qsort(BBCC **a, int n, int (*cmp)(BBCC**,BBCC**))
 	if ((s = a+n-1-pd)>0) { for(r=0;r<s;r++) swap(pc+r, a+n-s+r); }	    
 
 	CLG_DEBUGIF(8) {
-	  VG_(printf)("   PV BB %p, ", bb_addr((*pv)->bb));
+	  VG_(printf)("   PV BB %#lx, ", bb_addr((*pv)->bb));
 	    CLG_(print_cxt)(9, (*pv)->cxt, (*pv)->rec_index);
 
 	    s = pb-pa+1;
-	    VG_(printf)("    Lower %d - %d:\n", a-qsort_start, a+s-1-qsort_start);
+	    VG_(printf)("    Lower %ld - %ld:\n",
+                        a-qsort_start + 0L,
+                        a+s-1-qsort_start + 0L);
 	    for (r=0;r<s;r++) {
 		pm = a+r;
-		VG_(printf)("     %3d BB %p, ", 
-			    pm-qsort_start,bb_addr((*pm)->bb));
+		VG_(printf)("     %3ld BB %#lx, ",
+			    pm-qsort_start + 0L,
+                            bb_addr((*pm)->bb));
 		CLG_(print_cxt)(9, (*pm)->cxt, (*pm)->rec_index);
 	    }
 
 	    s = pd-pc+1;
-	    VG_(printf)("    Upper %d - %d:\n", 
-			a+n-s-qsort_start, a+n-1-qsort_start);
+	    VG_(printf)("    Upper %ld - %ld:\n",
+			a+n-s-qsort_start + 0L,
+                        a+n-1-qsort_start + 0L);
 	    for (r=0;r<s;r++) {
 		pm = a+n-s+r;
-		VG_(printf)("     %3d BB %p, ", 
-			    pm-qsort_start,bb_addr((*pm)->bb));
+		VG_(printf)("     %3ld BB %#lx, ",
+			    pm-qsort_start + 0L,
+                            bb_addr((*pm)->bb));
 		CLG_(print_cxt)(9, (*pm)->cxt, (*pm)->rec_index);
 	    }
 	}
@@ -1212,7 +1223,8 @@ BBCC** prepare_dump(void)
 
     /* allocate bbcc array, insert BBCCs and sort */
     prepare_ptr = array =
-      (BBCC**) CLG_MALLOC((prepare_count+1) * sizeof(BBCC*));    
+      (BBCC**) CLG_MALLOC("cl.dump.pd.1",
+                          (prepare_count+1) * sizeof(BBCC*));    
 
     CLG_(forall_bbccs)(hash_addPtr);
 
@@ -1239,7 +1251,7 @@ BBCC** prepare_dump(void)
 
 
 
-static void fprint_cost_ln(int fd, Char* prefix,
+static void fprint_cost_ln(int fd, const HChar* prefix,
 			   EventMapping* em, ULong* cost)
 {
     int p;
@@ -1247,17 +1259,17 @@ static void fprint_cost_ln(int fd, Char* prefix,
     p = VG_(sprintf)(outbuf, "%s", prefix);
     p += CLG_(sprint_mappingcost)(outbuf + p, em, cost);
     VG_(sprintf)(outbuf + p, "\n");
-    my_fwrite(fd, (void*)outbuf, VG_(strlen)(outbuf));
+    my_fwrite(fd, outbuf, VG_(strlen)(outbuf));
 }
 
 static ULong bbs_done = 0;
-static Char* filename = 0;
+static HChar* filename = 0;
 
 static
 void file_err(void)
 {
    VG_(message)(Vg_UserMsg,
-                "Error: can not open cache simulation output file `%s'",
+                "Error: can not open cache simulation output file `%s'\n",
                 filename );
    VG_(exit)(1);
 }
@@ -1271,7 +1283,7 @@ void file_err(void)
  *
  * Returns the file descriptor, and -1 on error (no write permission)
  */
-static int new_dumpfile(Char buf[BUF_LEN], int tid, Char* trigger)
+static int new_dumpfile(HChar buf[BUF_LEN], int tid, const HChar* trigger)
 {
     Bool appending = False;
     int i, fd;
@@ -1288,27 +1300,27 @@ static int new_dumpfile(Char buf[BUF_LEN], int tid, Char* trigger)
 	    i += VG_(sprintf)(filename+i, ".%d", out_counter);
 
 	if (CLG_(clo).separate_threads)
-	    i += VG_(sprintf)(filename+i, "-%02d", tid);
+	    VG_(sprintf)(filename+i, "-%02d", tid);
 
 	res = VG_(open)(filename, VKI_O_WRONLY|VKI_O_TRUNC, 0);
     }
     else {
 	VG_(sprintf)(filename, "%s", out_file);
         res = VG_(open)(filename, VKI_O_WRONLY|VKI_O_APPEND, 0);
-	if (!res.isError && out_counter>1)
+	if (!sr_isError(res) && out_counter>1)
 	    appending = True;
     }
 
-    if (res.isError) {
+    if (sr_isError(res)) {
 	res = VG_(open)(filename, VKI_O_CREAT|VKI_O_WRONLY,
 			VKI_S_IRUSR|VKI_S_IWUSR);
-	if (res.isError) {
+	if (sr_isError(res)) {
 	    /* If the file can not be opened for whatever reason (conflict
 	       between multiple supervised processes?), give up now. */
 	    file_err();
 	}
     }
-    fd = (Int) res.res;
+    fd = (Int) sr_Res(res);
 
     CLG_DEBUG(2, "  new_dumpfile '%s'\n", filename);
 
@@ -1319,27 +1331,27 @@ static int new_dumpfile(Char buf[BUF_LEN], int tid, Char* trigger)
     if (!appending) {
 	/* version */
 	VG_(sprintf)(buf, "version: 1\n");
-	my_fwrite(fd, (void*)buf, VG_(strlen)(buf));
+	my_fwrite(fd, buf, VG_(strlen)(buf));
 
 	/* creator */
 	VG_(sprintf)(buf, "creator: callgrind-" VERSION "\n");
-	my_fwrite(fd, (void*)buf, VG_(strlen)(buf));
+	my_fwrite(fd, buf, VG_(strlen)(buf));
 
 	/* "pid:" line */
 	VG_(sprintf)(buf, "pid: %d\n", VG_(getpid)());
-	my_fwrite(fd, (void*)buf, VG_(strlen)(buf));
+	my_fwrite(fd, buf, VG_(strlen)(buf));
 
 	/* "cmd:" line */
 	VG_(strcpy)(buf, "cmd: ");
-	my_fwrite(fd, (void*)buf, VG_(strlen)(buf));
-	my_fwrite(fd, (void*)cmdbuf, VG_(strlen)(cmdbuf));
+	my_fwrite(fd, buf, VG_(strlen)(buf));
+	my_fwrite(fd, cmdbuf, VG_(strlen)(cmdbuf));
     }
 
     VG_(sprintf)(buf, "\npart: %d\n", out_counter);
-    my_fwrite(fd, (void*)buf, VG_(strlen)(buf));
+    my_fwrite(fd, buf, VG_(strlen)(buf));
     if (CLG_(clo).separate_threads) {
 	VG_(sprintf)(buf, "thread: %d\n", tid);
-	my_fwrite(fd, (void*)buf, VG_(strlen)(buf));
+	my_fwrite(fd, buf, VG_(strlen)(buf));
     }
 
     /* "desc:" lines */
@@ -1350,36 +1362,36 @@ static int new_dumpfile(Char buf[BUF_LEN], int tid, Char* trigger)
 	/* Global options changing the tracing behaviour */
 	VG_(sprintf)(buf, "\ndesc: Option: --skip-plt=%s\n",
 		     CLG_(clo).skip_plt ? "yes" : "no");
-	my_fwrite(fd, (void*)buf, VG_(strlen)(buf));
+	my_fwrite(fd, buf, VG_(strlen)(buf));
 	VG_(sprintf)(buf, "desc: Option: --collect-jumps=%s\n",
 		     CLG_(clo).collect_jumps ? "yes" : "no");
-	my_fwrite(fd, (void*)buf, VG_(strlen)(buf));
+	my_fwrite(fd, buf, VG_(strlen)(buf));
 	VG_(sprintf)(buf, "desc: Option: --separate-recs=%d\n",
 		     CLG_(clo).separate_recursions);
-	my_fwrite(fd, (void*)buf, VG_(strlen)(buf));
+	my_fwrite(fd, buf, VG_(strlen)(buf));
 	VG_(sprintf)(buf, "desc: Option: --separate-callers=%d\n",
 		     CLG_(clo).separate_callers);
-	my_fwrite(fd, (void*)buf, VG_(strlen)(buf));
+	my_fwrite(fd, buf, VG_(strlen)(buf));
 
 	VG_(sprintf)(buf, "desc: Option: --dump-bbs=%s\n",
 		     CLG_(clo).dump_bbs ? "yes" : "no");
-	my_fwrite(fd, (void*)buf, VG_(strlen)(buf));
+	my_fwrite(fd, buf, VG_(strlen)(buf));
 	VG_(sprintf)(buf, "desc: Option: --separate-threads=%s\n",
 		     CLG_(clo).separate_threads ? "yes" : "no");
-	my_fwrite(fd, (void*)buf, VG_(strlen)(buf));
+	my_fwrite(fd, buf, VG_(strlen)(buf));
 #endif
 
 	(*CLG_(cachesim).getdesc)(buf);
-	my_fwrite(fd, (void*)buf, VG_(strlen)(buf));
+	my_fwrite(fd, buf, VG_(strlen)(buf));
     }
 
     VG_(sprintf)(buf, "\ndesc: Timerange: Basic block %llu - %llu\n",
 		 bbs_done, CLG_(stat).bb_executions);
 
-    my_fwrite(fd, (void*)buf, VG_(strlen)(buf));
+    my_fwrite(fd, buf, VG_(strlen)(buf));
     VG_(sprintf)(buf, "desc: Trigger: %s\n",
-		 trigger ? trigger : (Char*)"Program termination");
-    my_fwrite(fd, (void*)buf, VG_(strlen)(buf));
+		 trigger ? trigger : "Program termination");
+    my_fwrite(fd, buf, VG_(strlen)(buf));
 
 #if 0
    /* Output function specific config
@@ -1389,27 +1401,27 @@ static int new_dumpfile(Char buf[BUF_LEN], int tid, Char* trigger)
        while (fnc) {
 	   if (fnc->skip) {
 	       VG_(sprintf)(buf, "desc: Option: --fn-skip=%s\n", fnc->name);
-	       my_fwrite(fd, (void*)buf, VG_(strlen)(buf));
+	       my_fwrite(fd, buf, VG_(strlen)(buf));
 	   }
 	   if (fnc->dump_at_enter) {
 	       VG_(sprintf)(buf, "desc: Option: --fn-dump-at-enter=%s\n",
 			    fnc->name);
-	       my_fwrite(fd, (void*)buf, VG_(strlen)(buf));
+	       my_fwrite(fd, buf, VG_(strlen)(buf));
 	   }   
 	   if (fnc->dump_at_leave) {
 	       VG_(sprintf)(buf, "desc: Option: --fn-dump-at-leave=%s\n",
 			    fnc->name);
-	       my_fwrite(fd, (void*)buf, VG_(strlen)(buf));
+	       my_fwrite(fd, buf, VG_(strlen)(buf));
 	   }
 	   if (fnc->separate_callers != CLG_(clo).separate_callers) {
 	       VG_(sprintf)(buf, "desc: Option: --separate-callers%d=%s\n",
 			    fnc->separate_callers, fnc->name);
-	       my_fwrite(fd, (void*)buf, VG_(strlen)(buf));
+	       my_fwrite(fd, buf, VG_(strlen)(buf));
 	   }   
 	   if (fnc->separate_recursions != CLG_(clo).separate_recursions) {
 	       VG_(sprintf)(buf, "desc: Option: --separate-recs%d=%s\n",
 			    fnc->separate_recursions, fnc->name);
-	       my_fwrite(fd, (void*)buf, VG_(strlen)(buf));
+	       my_fwrite(fd, buf, VG_(strlen)(buf));
 	   }   
 	   fnc = fnc->next;
        }
@@ -1421,12 +1433,12 @@ static int new_dumpfile(Char buf[BUF_LEN], int tid, Char* trigger)
 		CLG_(clo).dump_instr ? " instr" : "",
 		CLG_(clo).dump_bb    ? " bb" : "",
 		CLG_(clo).dump_line  ? " line" : "");
-   my_fwrite(fd, (void*)buf, VG_(strlen)(buf));
+   my_fwrite(fd, buf, VG_(strlen)(buf));
 
    /* "events:" line */
    i = VG_(sprintf)(buf, "events: ");
    CLG_(sprint_eventmapping)(buf+i, CLG_(dumpmap));
-   my_fwrite(fd, (void*)buf, VG_(strlen)(buf));
+   my_fwrite(fd, buf, VG_(strlen)(buf));
    my_fwrite(fd, "\n", 1);
 
    /* summary lines */
@@ -1459,13 +1471,13 @@ static int new_dumpfile(Char buf[BUF_LEN], int tid, Char* trigger)
    my_fwrite(fd, "\n\n",2);
 
    if (VG_(clo_verbosity) > 1)
-       VG_(message)(Vg_DebugMsg, "Dump to %s", filename);
+       VG_(message)(Vg_DebugMsg, "Dump to %s\n", filename);
 
    return fd;
 }
 
 
-static void close_dumpfile(Char buf[BUF_LEN], int fd, int tid)
+static void close_dumpfile(int fd)
 {
     if (fd <0) return;
 
@@ -1481,7 +1493,7 @@ static void close_dumpfile(Char buf[BUF_LEN], int fd, int tid)
     if (filename[0] == '.') {
 	if (-1 == VG_(rename) (filename, filename+1)) {
 	    /* Can not rename to correct file name: give out warning */
-	    VG_(message)(Vg_DebugMsg, "Warning: Can not rename .%s to %s",
+	    VG_(message)(Vg_DebugMsg, "Warning: Can not rename .%s to %s\n",
 			 filename, filename);
        }
    }
@@ -1491,8 +1503,8 @@ static void close_dumpfile(Char buf[BUF_LEN], int fd, int tid)
 /* Helper for print_bbccs */
 
 static Int   print_fd;
-static Char* print_trigger;
-static Char  print_buf[BUF_LEN];
+static const HChar* print_trigger;
+static HChar print_buf[BUF_LEN];
 
 static void print_bbccs_of_thread(thread_info* ti)
 {
@@ -1529,7 +1541,7 @@ static void print_bbccs_of_thread(thread_info* ti)
 	/* switch back to file of function */
 	VG_(sprintf)(print_buf, "fe=");
 	print_file(print_buf+3, lastFnPos.cxt->fn[0]->file);
-	my_fwrite(print_fd, (void*)print_buf, VG_(strlen)(print_buf));
+	my_fwrite(print_fd, print_buf, VG_(strlen)(print_buf));
       }
       my_fwrite(print_fd, "\n", 1);
     }
@@ -1549,7 +1561,7 @@ static void print_bbccs_of_thread(thread_info* ti)
 	/* FIXME: Specify Object of BB if different to object of fn */
 	int i, pos = 0;
 	ULong ecounter = (*p)->ecounter_sum;
-	pos = VG_(sprintf)(print_buf, "bb=%p ", (*p)->bb->offset);
+	pos = VG_(sprintf)(print_buf, "bb=%#lx ", (*p)->bb->offset);
 	for(i = 0; i<(*p)->bb->cjmp_count;i++) {
 	    pos += VG_(sprintf)(print_buf+pos, "%d %llu ", 
 				(*p)->bb->jmp[i].instr,
@@ -1559,7 +1571,7 @@ static void print_bbccs_of_thread(thread_info* ti)
 	VG_(sprintf)(print_buf+pos, "%d %llu\n", 
 		     (*p)->bb->instr_count,
 		     ecounter);
-	my_fwrite(print_fd, (void*)print_buf, VG_(strlen)(print_buf));
+	my_fwrite(print_fd, print_buf, VG_(strlen)(print_buf));
     }
     
     fprint_bbcc(print_fd, *p, &lastAPos);
@@ -1567,7 +1579,7 @@ static void print_bbccs_of_thread(thread_info* ti)
     p++;
   }
 
-  close_dumpfile(print_buf, print_fd, CLG_(current_tid));
+  close_dumpfile(print_fd);
   if (array) VG_(free)(array);
   
   /* set counters of last dump */
@@ -1578,7 +1590,7 @@ static void print_bbccs_of_thread(thread_info* ti)
 }
 
 
-static void print_bbccs(Char* trigger, Bool only_current_thread)
+static void print_bbccs(const HChar* trigger, Bool only_current_thread)
 {
   init_dump_array();
   init_debug_cache();
@@ -1603,15 +1615,17 @@ static void print_bbccs(Char* trigger, Bool only_current_thread)
 }
 
 
-void CLG_(dump_profile)(Char* trigger, Bool only_current_thread)
+void CLG_(dump_profile)(const HChar* trigger, Bool only_current_thread)
 {
    CLG_DEBUG(2, "+ dump_profile(Trigger '%s')\n",
-	    trigger ? trigger : (Char*)"Prg.Term.");
+	    trigger ? trigger : "Prg.Term.");
+
+   CLG_(init_dumps)();
 
    if (VG_(clo_verbosity) > 1)
-       VG_(message)(Vg_DebugMsg, "Start dumping at BB %llu (%s)...",
+       VG_(message)(Vg_DebugMsg, "Start dumping at BB %llu (%s)...\n",
 		    CLG_(stat).bb_executions,
-		    trigger ? trigger : (Char*)"Prg.Term.");
+		    trigger ? trigger : "Prg.Term.");
 
    out_counter++;
 
@@ -1620,19 +1634,22 @@ void CLG_(dump_profile)(Char* trigger, Bool only_current_thread)
    bbs_done = CLG_(stat).bb_executions++;
 
    if (VG_(clo_verbosity) > 1)
-     VG_(message)(Vg_DebugMsg, "Dumping done.");
+     VG_(message)(Vg_DebugMsg, "Dumping done.\n");
 }
 
-/* copy command to cmd buffer (could change) */
+/* Copy command to cmd buffer. We want to original command line
+ * (can change at runtime)
+ */
 static
 void init_cmdbuf(void)
 {
   Int i,j,size = 0;
   HChar* argv;
 
-#if VG_CORE_INTERFACE_VERSION > 8
-  if (VG_(args_the_exename))
+  if (VG_(args_the_exename)) {
+      CLG_ASSERT( VG_(strlen)( VG_(args_the_exename) ) < BUF_LEN-1);
       size = VG_(sprintf)(cmdbuf, " %s", VG_(args_the_exename));
+  }
 
   for(i = 0; i < VG_(sizeXA)( VG_(args_for_client) ); i++) {
       argv = * (HChar**) VG_(indexXA)( VG_(args_for_client), i );
@@ -1641,17 +1658,8 @@ void init_cmdbuf(void)
       for(j=0;argv[j]!=0;j++)
 	  if (size < BUF_LEN) cmdbuf[size++] = argv[j];
   }
-#else
-  for(i = 0; i < VG_(client_argc); i++) {
-    argv = VG_(client_argv)[i];
-    if (!argv) continue;
-    if ((size>0) && (size < BUF_LEN)) cmdbuf[size++] = ' ';
-    for(j=0;argv[j]!=0;j++)
-      if (size < BUF_LEN) cmdbuf[size++] = argv[j];
-  }
-#endif
 
-  if (size == BUF_LEN) size--;
+  if (size >= BUF_LEN) size = BUF_LEN-1;
   cmdbuf[size] = 0;
 }
 
@@ -1665,15 +1673,35 @@ void init_cmdbuf(void)
  * <out_file> always starts with a full absolute path.
  * If the output format string represents a relative path, the current
  * working directory at program start is used.
+ *
+ * This function has to be called every time a profile dump is generated
+ * to be able to react on PID changes.
  */
 void CLG_(init_dumps)()
 {
    Int lastSlash, i;
    SysRes res;
 
+   static int thisPID = 0;
+   int currentPID = VG_(getpid)();
+   if (currentPID == thisPID) {
+       /* already initialized, and no PID change */
+       CLG_ASSERT(out_file != 0);
+       return;
+   }
+   thisPID = currentPID;
+   
    if (!CLG_(clo).out_format)
      CLG_(clo).out_format = DEFAULT_OUTFORMAT;
 
+   /* If a file name was already set, clean up before */
+   if (out_file) {
+       VG_(free)(out_file);
+       VG_(free)(out_directory);
+       VG_(free)(filename);
+       out_counter = 0;
+   }
+
    // Setup output filename.
    out_file =
        VG_(expand_file_name)("--callgrind-out-file", CLG_(clo).out_format);
@@ -1687,12 +1715,13 @@ void CLG_(init_dumps)()
        i++;
    }
    i = lastSlash;
-   out_directory = (Char*) CLG_MALLOC(i+1);
+   out_directory = (HChar*) CLG_MALLOC("cl.dump.init_dumps.1", i+1);
    VG_(strncpy)(out_directory, out_file, i);
    out_directory[i] = 0;
 
    /* allocate space big enough for final filenames */
-   filename = (Char*) CLG_MALLOC(VG_(strlen)(out_file)+32);
+   filename = (HChar*) CLG_MALLOC("cl.dump.init_dumps.2",
+                                 VG_(strlen)(out_file)+32);
    CLG_ASSERT(filename != 0);
        
    /* Make sure the output base file can be written.
@@ -1703,16 +1732,17 @@ void CLG_(init_dumps)()
     */ 
     VG_(strcpy)(filename, out_file);
     res = VG_(open)(filename, VKI_O_WRONLY|VKI_O_TRUNC, 0);
-    if (res.isError) { 
+    if (sr_isError(res)) { 
 	res = VG_(open)(filename, VKI_O_CREAT|VKI_O_WRONLY,
 		       VKI_S_IRUSR|VKI_S_IWUSR);
-	if (res.isError) {
+	if (sr_isError(res)) {
 	    file_err(); 
 	}
     }
-    if (!res.isError) VG_(close)( (Int)res.res );
+    if (!sr_isError(res)) VG_(close)( (Int)sr_Res(res) );
 
-    init_cmdbuf();
+    if (!dumps_initialized)
+	init_cmdbuf();
 
     dumps_initialized = True;
 }
