@@ -6,7 +6,7 @@
 /*
    This file is part of Callgrind, a Valgrind tool for call tracing.
 
-   Copyright (C) 2002-2007, Josef Weidendorfer (Josef.Weidendorfer@gmx.de)
+   Copyright (C) 2002-2013, Josef Weidendorfer (Josef.Weidendorfer@gmx.de)
 
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
@@ -35,7 +35,79 @@ static fn_array current_fn_active;
 static Addr runtime_resolve_addr = 0;
 static int  runtime_resolve_length = 0;
 
-/* _ld_runtime_resolve, located in needs special handling:
+// a code pattern is a list of tuples (start offset, length)
+struct chunk_t { int start, len; };
+struct pattern
+{
+    const HChar* name;
+    int len;
+    struct chunk_t chunk[];
+};
+
+/* Scan for a pattern in the code of an ELF object.
+ * If found, return true and set runtime_resolve_{addr,length}
+ */
+__attribute__((unused))    // Possibly;  depends on the platform.
+static Bool check_code(obj_node* obj,
+                       UChar code[], struct pattern* pat)
+{
+    Bool found;
+    Addr addr, end;
+    int chunk, start, len;
+
+    /* first chunk of pattern should always start at offset 0 and
+     * have at least 3 bytes */
+    CLG_ASSERT((pat->chunk[0].start == 0) && (pat->chunk[0].len >2));
+    
+    CLG_DEBUG(1, "check_code: %s, pattern %s, check %d bytes of [%x %x %x...]\n",
+              obj->name, pat->name, pat->chunk[0].len, code[0], code[1], code[2]);
+
+    end = obj->start + obj->size - pat->len;
+    addr = obj->start;
+    while(addr < end) {
+	found = (VG_(memcmp)( (void*)addr, code, pat->chunk[0].len) == 0);
+
+        if (found) {
+	    chunk = 1;
+	    while(1) {		
+		start = pat->chunk[chunk].start;
+		len   = pat->chunk[chunk].len;
+		if (len == 0) break;
+
+		CLG_ASSERT(len >2);
+                CLG_DEBUG(1, " found chunk %d at %#lx, checking %d bytes "
+                             "of [%x %x %x...]\n",
+                          chunk-1, addr - obj->start, len,
+			  code[start], code[start+1], code[start+2]);
+
+                if (VG_(memcmp)( (void*)(addr+start), code+start, len) != 0) {
+                    found = False;
+                    break;
+                }
+		chunk++;
+	    }
+
+            if (found) {
+		CLG_DEBUG(1, "found at offset %#lx.\n", addr - obj->start);
+		if (VG_(clo_verbosity) > 1)
+		    VG_(message)(Vg_DebugMsg, "Found runtime_resolve (%s): "
+                                              "%s +%#lx=%#lx, length %d\n",
+				 pat->name, obj->name + obj->last_slash_pos,
+				 addr - obj->start, addr, pat->len);
+		    
+		runtime_resolve_addr   = addr;
+		runtime_resolve_length = pat->len;
+		return True;
+	    }
+        }
+        addr++;
+    }
+    CLG_DEBUG(1, " found nothing.\n");
+    return False;
+}
+
+
+/* _ld_runtime_resolve, located in ld.so, needs special handling:
  * The jump at end into the resolved function should not be
  * represented as a call (as usually done in callgrind with jumps),
  * but as a return + call. Otherwise, the repeated existance of
@@ -43,27 +115,38 @@ static int  runtime_resolve_length = 0;
  * making the profile almost worthless.
  *
  * If ld.so is stripped, the symbol will not appear. But as this
- * function is handcrafted assembler, we search for it...
+ * function is handcrafted assembler, we search for it.
  *
- * Returns 0 if code not found, otherwise start address
+ * We stop if the ELF object name does not seem to be the runtime linker
  */
-static void search_runtime_resolve(obj_node* obj)
+static Bool search_runtime_resolve(obj_node* obj)
 {
-    /* We do not check target address of <fixup>, therefore we have >1 ranges.
-     * We use a tuple sequence (offset,length) into the code array for this
-     */
-
 #if defined(VGP_x86_linux)
-    /* Check ranges [0-11], [16-23] */
-    static int  code_offsets[] = { 0, 12, 16, 8, 24, 0 };
-    static unsigned char code[] = {
+    static UChar code[] = {
 	/* 0*/ 0x50, 0x51, 0x52, 0x8b, 0x54, 0x24, 0x10, 0x8b,
 	/* 8*/ 0x44, 0x24, 0x0c, 0xe8, 0x70, 0x01, 0x00, 0x00,
 	/*16*/ 0x5a, 0x59, 0x87, 0x04, 0x24, 0xc2, 0x08, 0x00 };
-#else
+    /* Check ranges [0-11] and [16-23] ([12-15] is an absolute address) */
+    static struct pattern pat = {
+	"x86-def", 24, {{ 0,12 }, { 16,8 }, { 24,0}} };
+
+    /* Pattern for glibc-2.8 on OpenSuse11.0 */
+    static UChar code_28[] = {
+	/* 0*/ 0x50, 0x51, 0x52, 0x8b, 0x54, 0x24, 0x10, 0x8b,
+	/* 8*/ 0x44, 0x24, 0x0c, 0xe8, 0x70, 0x01, 0x00, 0x00,
+	/*16*/ 0x5a, 0x8b, 0x0c, 0x24, 0x89, 0x04, 0x24, 0x8b,
+	/*24*/ 0x44, 0x24, 0x04, 0xc2, 0x0c, 0x00 };
+    static struct pattern pat_28 = {
+	"x86-glibc2.8", 30, {{ 0,12 }, { 16,14 }, { 30,0}} };
+
+    if (VG_(strncmp)(obj->name, "/lib/ld", 7) != 0) return False;
+    if (check_code(obj, code, &pat)) return True;
+    if (check_code(obj, code_28, &pat_28)) return True;
+    return False;
+#endif
+
 #if defined(VGP_ppc32_linux)
-    static int  code_offsets[] = {0, 65, 68, 64, 132, 0 };
-    static unsigned char code[] = {
+    static UChar code[] = {
 	/* 0*/ 0x94, 0x21, 0xff, 0xc0, 0x90, 0x01, 0x00, 0x0c,
 	/* 8*/ 0x90, 0x61, 0x00, 0x10, 0x90, 0x81, 0x00, 0x14,
 	/*16*/ 0x7d, 0x83, 0x63, 0x78, 0x90, 0xa1, 0x00, 0x18,
@@ -81,11 +164,15 @@ static void search_runtime_resolve(obj_node* obj)
 	/*112*/0x80, 0x81, 0x00, 0x14, 0x80, 0x61, 0x00, 0x10,
 	/*120*/0x80, 0x01, 0x00, 0x0c, 0x38, 0x21, 0x00, 0x40,
 	/*128*/0x4e, 0x80, 0x04, 0x20 };
-#else
+    static struct pattern pat = {
+	"ppc32-def", 132, {{ 0,65 }, { 68,64 }, { 132,0 }} };
+
+    if (VG_(strncmp)(obj->name, "/lib/ld", 7) != 0) return False;
+    return check_code(obj, code, &pat);
+#endif
+
 #if defined(VGP_amd64_linux)
-    /* x86_64 */
-    static int  code_offsets[] = {0, 62, 66, 44, 110, 0 };
-    static unsigned char code[] = {
+    static UChar code[] = {
 	/* 0*/ 0x48, 0x83, 0xec, 0x38, 0x48, 0x89, 0x04, 0x24,
 	/* 8*/ 0x48, 0x89, 0x4c, 0x24, 0x08, 0x48, 0x89, 0x54, 0x24, 0x10,
 	/*18*/ 0x48, 0x89, 0x74, 0x24, 0x18, 0x48, 0x89, 0x7c, 0x24, 0x20,
@@ -98,61 +185,19 @@ static void search_runtime_resolve(obj_node* obj)
 	/*84*/ 0x48, 0x8b, 0x74, 0x24, 0x18, 0x48, 0x8b, 0x54, 0x24, 0x10,
 	/*94*/ 0x48, 0x8b, 0x4c, 0x24, 0x08, 0x48, 0x8b, 0x04, 0x24,
 	/*103*/0x48, 0x83, 0xc4, 0x48, 0x41, 0xff, 0xe3 };
-#else
-    /* Unknown platform, no check is done */
-    static int  code_offsets[] = {0, 0 };
-    static unsigned char code[] = { 0 };
-#endif
-#endif
-#endif
-    
-    int *range = &(code_offsets[0]), *r = 0;
-    Bool found = False;
-    Addr addr, end;
+    static struct pattern pat = {
+	"amd64-def", 110, {{ 0,62 }, { 66,44 }, { 110,0 }} };
 
-    /* Only search in libraries with a given name pattern */
     if ((VG_(strncmp)(obj->name, "/lib/ld", 7) != 0) &&
-	(VG_(strncmp)(obj->name, "/lib64/ld", 9) != 0)) return;
-    
-    CLG_DEBUG(1, "search_rs: Checking %d bytes of [%x %x %x...]\n",
-	      range[1], code[0], code[1], code[2]);
+	(VG_(strncmp)(obj->name, "/lib64/ld", 9) != 0)) return False;
+    return check_code(obj, code, &pat);
+#endif
 
-    end = obj->start + obj->size - range[1];
-    addr = obj->start;
-
-    if (range[1] == 0) return;
-
-    while(addr < end) {
-	if (VG_(memcmp)( (void*)addr, code, range[1]) == 0) {
-
-	    r = range + 2;
-	    found = True;
-	    while(r[1]) {
-		CLG_DEBUG(1, " [%p] Found! Checking %d bytes of [%x %x %x...]\n",
-			  addr, r[1], code[r[0]], code[r[0]+1], code[r[0]+2]);
-
-		if (VG_(memcmp)( (void*)(addr+r[0]), code+r[0], r[1]) != 0) {
-		    found = False;
-		    break;
-		}
-		r += 2;
-	    }
-	    if (found) break;
-	}
-	addr++;
-    }
-
-    if (!found || (r==0)) return;
-
-    if (VG_(clo_verbosity) > 1)
-	VG_(message)(Vg_DebugMsg, "Code check found runtime_resolve: %s +%p=%p, length %d",
-		     obj->name + obj->last_slash_pos,
-		     addr - obj->start, addr, r[0]);
-
-    runtime_resolve_addr   = addr;
-    runtime_resolve_length = r[0];
+    /* For other platforms, no patterns known */
+    return False;
 }
 
+
 /*------------------------------------------------------------*/
 /*--- Object/File/Function hash entry operations           ---*/
 /*------------------------------------------------------------*/
@@ -169,7 +214,7 @@ void CLG_(init_obj_table)()
 
 #define HASH_CONSTANT   256
 
-static UInt str_hash(const Char *s, UInt table_size)
+static UInt str_hash(const HChar *s, UInt table_size)
 {
     int hash_value = 0;
     for ( ; *s; s++)
@@ -178,47 +223,51 @@ static UInt str_hash(const Char *s, UInt table_size)
 }
 
 
-static Char* anonymous_obj = "???";
+static const HChar* anonymous_obj = "???";
 
 static __inline__ 
-obj_node* new_obj_node(SegInfo* si, obj_node* next)
+obj_node* new_obj_node(DebugInfo* di, obj_node* next)
 {
    Int i;
-   obj_node* new;
+   obj_node* obj;
 
-   new = (obj_node*) CLG_MALLOC(sizeof(obj_node));
-   new->name  = si ? VG_(strdup)( VG_(seginfo_filename)(si) )
-                     : anonymous_obj;
+   obj = (obj_node*) CLG_MALLOC("cl.fn.non.1", sizeof(obj_node));
+   obj->name  = di ? VG_(strdup)( "cl.fn.non.2",
+                                  VG_(DebugInfo_get_filename)(di) )
+                   : anonymous_obj;
    for (i = 0; i < N_FILE_ENTRIES; i++) {
-      new->files[i] = NULL;
+      obj->files[i] = NULL;
    }
    CLG_(stat).distinct_objs ++;
-   new->number  = CLG_(stat).distinct_objs;
-   new->start   = si ? VG_(seginfo_start)(si) : 0;
-   new->size    = si ? VG_(seginfo_size)(si) : 0;
-   new->offset  = si ? VG_(seginfo_sym_offset)(si) : 0;
-   new->next    = next;
+   obj->number  = CLG_(stat).distinct_objs;
+   /* JRS 2008 Feb 19: maybe rename .start/.size/.offset to
+      .text_avma/.text_size/.test_bias to make it clearer what these
+      fields really mean */
+   obj->start   = di ? VG_(DebugInfo_get_text_avma)(di) : 0;
+   obj->size    = di ? VG_(DebugInfo_get_text_size)(di) : 0;
+   obj->offset  = di ? VG_(DebugInfo_get_text_bias)(di) : 0;
+   obj->next    = next;
 
    // not only used for debug output (see static.c)
-   new->last_slash_pos = 0;
+   obj->last_slash_pos = 0;
    i = 0;
-   while(new->name[i]) {
-	if (new->name[i]=='/') new->last_slash_pos = i+1;
+   while(obj->name[i]) {
+	if (obj->name[i]=='/') obj->last_slash_pos = i+1;
 	i++;
    }
 
-   if (runtime_resolve_addr == 0) search_runtime_resolve(new);
-   
-   return new;
+   if (runtime_resolve_addr == 0) search_runtime_resolve(obj);
+
+   return obj;
 }
 
-obj_node* CLG_(get_obj_node)(SegInfo* si)
+obj_node* CLG_(get_obj_node)(DebugInfo* di)
 {
     obj_node*    curr_obj_node;
     UInt         objname_hash;
-    const UChar* obj_name;
+    const HChar* obj_name;
     
-    obj_name = si ? (Char*) VG_(seginfo_filename)(si) : anonymous_obj;
+    obj_name = di ? VG_(DebugInfo_get_filename)(di) : anonymous_obj;
 
     /* lookup in obj hash */
     objname_hash = str_hash(obj_name, N_OBJ_ENTRIES);
@@ -229,7 +278,7 @@ obj_node* CLG_(get_obj_node)(SegInfo* si)
     }
     if (NULL == curr_obj_node) {
 	obj_table[objname_hash] = curr_obj_node = 
-	    new_obj_node(si, obj_table[objname_hash]);
+	    new_obj_node(di, obj_table[objname_hash]);
     }
 
     return curr_obj_node;
@@ -237,25 +286,26 @@ obj_node* CLG_(get_obj_node)(SegInfo* si)
 
 
 static __inline__ 
-file_node* new_file_node(Char filename[FILENAME_LEN],
+file_node* new_file_node(HChar filename[FILENAME_LEN],
 			 obj_node* obj, file_node* next)
 {
   Int i;
-  file_node* new = (file_node*) CLG_MALLOC(sizeof(file_node));
-  new->name  = VG_(strdup)(filename);
+  file_node* file = (file_node*) CLG_MALLOC("cl.fn.nfn.1",
+                                           sizeof(file_node));
+  file->name  = VG_(strdup)("cl.fn.nfn.2", filename);
   for (i = 0; i < N_FN_ENTRIES; i++) {
-    new->fns[i] = NULL;
+    file->fns[i] = NULL;
   }
   CLG_(stat).distinct_files++;
-  new->number  = CLG_(stat).distinct_files;
-  new->obj     = obj;
-  new->next      = next;
-  return new;
+  file->number  = CLG_(stat).distinct_files;
+  file->obj     = obj;
+  file->next      = next;
+  return file;
 }
 
  
 file_node* CLG_(get_file_node)(obj_node* curr_obj_node,
-			      Char filename[FILENAME_LEN])
+                               HChar filename[FILENAME_LEN])
 {
     file_node* curr_file_node;
     UInt       filename_hash;
@@ -280,41 +330,42 @@ file_node* CLG_(get_file_node)(obj_node* curr_obj_node,
 static void resize_fn_array(void);
 
 static __inline__ 
-fn_node* new_fn_node(Char fnname[FILENAME_LEN],
+fn_node* new_fn_node(HChar fnname[FILENAME_LEN],
 		     file_node* file, fn_node* next)
 {
-    fn_node* new = (fn_node*) CLG_MALLOC(sizeof(fn_node));
-    new->name = VG_(strdup)(fnname);
+    fn_node* fn = (fn_node*) CLG_MALLOC("cl.fn.nfnnd.1",
+                                         sizeof(fn_node));
+    fn->name = VG_(strdup)("cl.fn.nfnnd.2", fnname);
 
     CLG_(stat).distinct_fns++;
-    new->number   = CLG_(stat).distinct_fns;
-    new->last_cxt = 0;
-    new->pure_cxt = 0;
-    new->file     = file;
-    new->next     = next;
+    fn->number   = CLG_(stat).distinct_fns;
+    fn->last_cxt = 0;
+    fn->pure_cxt = 0;
+    fn->file     = file;
+    fn->next     = next;
 
-    new->dump_before  = False;
-    new->dump_after   = False;
-    new->zero_before  = False;
-    new->toggle_collect = False;
-    new->skip         = False;
-    new->pop_on_jump  = CLG_(clo).pop_on_jump;
-    new->is_malloc    = False;
-    new->is_realloc   = False;
-    new->is_free      = False;
+    fn->dump_before  = False;
+    fn->dump_after   = False;
+    fn->zero_before  = False;
+    fn->toggle_collect = False;
+    fn->skip         = False;
+    fn->pop_on_jump  = CLG_(clo).pop_on_jump;
+    fn->is_malloc    = False;
+    fn->is_realloc   = False;
+    fn->is_free      = False;
 
-    new->group        = 0;
-    new->separate_callers    = CLG_(clo).separate_callers;
-    new->separate_recursions = CLG_(clo).separate_recursions;
+    fn->group        = 0;
+    fn->separate_callers    = CLG_(clo).separate_callers;
+    fn->separate_recursions = CLG_(clo).separate_recursions;
 
 #if CLG_ENABLE_DEBUG
-    new->verbosity    = -1;
+    fn->verbosity    = -1;
 #endif
 
     if (CLG_(stat).distinct_fns >= current_fn_active.size)
 	resize_fn_array();
 
-    return new;
+    return fn;
 }
 
 
@@ -323,7 +374,7 @@ fn_node* new_fn_node(Char fnname[FILENAME_LEN],
  */
 static
 fn_node* get_fn_node_infile(file_node* curr_file_node,
-			    Char fnname[FN_NAME_LEN])
+			    HChar fnname[FN_NAME_LEN])
 {
     fn_node* curr_fn_node;
     UInt     fnname_hash;
@@ -351,11 +402,11 @@ fn_node* get_fn_node_infile(file_node* curr_file_node,
  * Hash nodes are created if needed.
  */
 static __inline__
-fn_node* get_fn_node_inseg(SegInfo* si,
-			   Char filename[FILENAME_LEN],
-			   Char fnname[FN_NAME_LEN])
+fn_node* get_fn_node_inseg(DebugInfo* di,
+			   HChar filename[FILENAME_LEN],
+			   HChar fnname[FN_NAME_LEN])
 {
-  obj_node  *obj  = CLG_(get_obj_node)(si);
+  obj_node  *obj  = CLG_(get_obj_node)(di);
   file_node *file = CLG_(get_file_node)(obj, filename);
   fn_node   *fn   = get_fn_node_infile(file, fnname);
 
@@ -364,18 +415,18 @@ fn_node* get_fn_node_inseg(SegInfo* si,
 
 
 Bool CLG_(get_debug_info)(Addr instr_addr,
-			 Char file[FILENAME_LEN],
-			 Char fn_name[FN_NAME_LEN], UInt* line_num,
-			 SegInfo** pSegInfo)
+			 HChar file[FILENAME_LEN],
+			 HChar fn_name[FN_NAME_LEN], UInt* line_num,
+			 DebugInfo** pDebugInfo)
 {
   Bool found_file_line, found_fn, found_dirname, result = True;
-  Char dir[FILENAME_LEN];
+  HChar dir[FILENAME_LEN];
   UInt line;
   
-  CLG_DEBUG(6, "  + get_debug_info(%p)\n", instr_addr);
+  CLG_DEBUG(6, "  + get_debug_info(%#lx)\n", instr_addr);
 
-  if (pSegInfo) {
-      *pSegInfo = VG_(find_seginfo)(instr_addr);
+  if (pDebugInfo) {
+      *pDebugInfo = VG_(find_DebugInfo)(instr_addr);
 
       // for generated code in anonymous space, pSegInfo is 0
    }
@@ -418,11 +469,11 @@ Bool CLG_(get_debug_info)(Addr instr_addr,
      if (line_num) *line_num=0;
    }
 
-   CLG_DEBUG(6, "  - get_debug_info(%p): seg '%s', fn %s\n",
+   CLG_DEBUG(6, "  - get_debug_info(%#lx): seg '%s', fn %s\n",
 	    instr_addr,
-	    !pSegInfo   ? (const UChar*)"-" :
-	    (*pSegInfo) ? VG_(seginfo_filename)(*pSegInfo) :
-	    (const UChar*)"(None)",
+	    !pDebugInfo   ? "-" :
+	    (*pDebugInfo) ? VG_(DebugInfo_get_filename)(*pDebugInfo) :
+	    "(None)",
 	    fn_name);
 
   return result;
@@ -437,31 +488,31 @@ static BB* exit_bb = 0;
  */
 fn_node* CLG_(get_fn_node)(BB* bb)
 {
-    Char       filename[FILENAME_LEN], fnname[FN_NAME_LEN];
-    SegInfo*   si;
+    HChar      filename[FILENAME_LEN], fnname[FN_NAME_LEN];
+    DebugInfo* di;
     UInt       line_num;
     fn_node*   fn;
 
     /* fn from debug info is idempotent for a BB */
     if (bb->fn) return bb->fn;
 
-    CLG_DEBUG(3,"+ get_fn_node(BB %p)\n", bb_addr(bb));
+    CLG_DEBUG(3,"+ get_fn_node(BB %#lx)\n", bb_addr(bb));
 
     /* get function/file name, line number and object of
      * the BB according to debug information
      */
     CLG_(get_debug_info)(bb_addr(bb),
-			filename, fnname, &line_num, &si);
+			filename, fnname, &line_num, &di);
 
     if (0 == VG_(strcmp)(fnname, "???")) {
 	int p;
 
 	/* Use address as found in library */
 	if (sizeof(Addr) == 4)
-	    p = VG_(sprintf)(fnname, "%08p", bb->offset);
+	    p = VG_(sprintf)(fnname, "%#08lx", bb->offset);
 	else 	    
 	    // 64bit address
-	    p = VG_(sprintf)(fnname, "%016p", bb->offset);
+	    p = VG_(sprintf)(fnname, "%#016lx", bb->offset);
 
 	VG_(sprintf)(fnname+p, "%s", 
 		     (bb->sect_kind == Vg_SectData) ? " [Data]" :
@@ -481,7 +532,7 @@ fn_node* CLG_(get_fn_node)(BB* bb)
     if (0 == VG_(strcmp)(fnname, "vgPlain___libc_freeres_wrapper")
 	&& exit_bb) {
       CLG_(get_debug_info)(bb_addr(exit_bb),
-			  filename, fnname, &line_num, &si);
+			  filename, fnname, &line_num, &di);
 	
 	CLG_DEBUG(1, "__libc_freeres_wrapper renamed to _exit\n");
     }
@@ -496,7 +547,7 @@ fn_node* CLG_(get_fn_node)(BB* bb)
     }
 
     /* get fn_node struct for this function */
-    fn = get_fn_node_inseg( si, filename, fnname);
+    fn = get_fn_node_inseg( di, filename, fnname);
 
     /* if this is the 1st time the function is seen,
      * some attributes are set */
@@ -518,7 +569,8 @@ fn_node* CLG_(get_fn_node)(BB* bb)
 	  fn->pop_on_jump = True;
 
 	  if (VG_(clo_verbosity) > 1)
-	      VG_(message)(Vg_DebugMsg, "Symbol match: found runtime_resolve: %s +%p=%p",
+	      VG_(message)(Vg_DebugMsg, "Symbol match: found runtime_resolve:"
+                                        " %s +%#lx=%#lx\n",
 		      bb->obj->name + bb->obj->last_slash_pos,
 		      bb->offset, bb_addr(bb));
       }
@@ -536,7 +588,7 @@ fn_node* CLG_(get_fn_node)(BB* bb)
     bb->fn   = fn;
     bb->line = line_num;
 
-    CLG_DEBUG(3,"- get_fn_node(BB %p): %s (in %s:%u)\n",
+    CLG_DEBUG(3,"- get_fn_node(BB %#lx): %s (in %s:%u)\n",
 	     bb_addr(bb), fnname, filename, line_num);
 
     return fn;
@@ -571,7 +623,8 @@ void CLG_(init_fn_array)(fn_array* a)
   if (a->size <= CLG_(stat).distinct_fns)
     a->size = CLG_(stat).distinct_fns+1;
   
-  a->array = (UInt*) CLG_MALLOC(a->size * sizeof(UInt));
+  a->array = (UInt*) CLG_MALLOC("cl.fn.gfe.1",
+                                a->size * sizeof(UInt));
   for(i=0;i<a->size;i++)
     a->array[i] = 0;
 }
@@ -605,7 +658,7 @@ void CLG_(set_current_fn_array)(fn_array* a)
  */
 static void resize_fn_array(void)
 {
-    UInt* new;
+    UInt* new_array;
     Int i, newsize;
 
     newsize = current_fn_active.size;
@@ -614,15 +667,15 @@ static void resize_fn_array(void)
     CLG_DEBUG(0, "Resize fn_active_array: %d => %d\n",
 	     current_fn_active.size, newsize);
 
-    new = (UInt*) CLG_MALLOC(newsize * sizeof(UInt));
+    new_array = (UInt*) CLG_MALLOC("cl.fn.rfa.1", newsize * sizeof(UInt));
     for(i=0;i<current_fn_active.size;i++)
-      new[i] = current_fn_active.array[i];
+      new_array[i] = current_fn_active.array[i];
     while(i<newsize)
-	new[i++] = 0;
+	new_array[i++] = 0;
 
     VG_(free)(current_fn_active.array);
     current_fn_active.size = newsize;
-    current_fn_active.array = new;
+    current_fn_active.array = new_array;
     CLG_(stat).fn_array_resizes++;
 }
 
