@@ -7,7 +7,7 @@
    This file is part of Valgrind, a dynamic binary instrumentation
    framework.
 
-   Copyright (C) 2008-2010 OpenWorks LLP
+   Copyright (C) 2008-2013 OpenWorks LLP
       info@open-works.co.uk
 
    This program is free software; you can redistribute it and/or
@@ -54,9 +54,10 @@ Bool ML_(TyEnt__is_type)( TyEnt* te )
       case Te_EMPTY: case Te_INDIR: case Te_UNKNOWN: 
       case Te_Atom:  case Te_Field: case Te_Bound:
          return False;
-      case Te_TyBase:   case Te_TyPorR: case Te_TyTyDef:
-      case Te_TyStOrUn: case Te_TyEnum: case Te_TyArray:
-      case Te_TyFn:     case Te_TyQual: case Te_TyVoid:
+      case Te_TyBase:   case Te_TyPtr:     case Te_TyRef:
+      case Te_TyPtrMbr: case Te_TyRvalRef: case Te_TyTyDef:
+      case Te_TyStOrUn: case Te_TyEnum:    case Te_TyArray:
+      case Te_TyFn:     case Te_TyQual:    case Te_TyVoid:
          return True;
       default:
          vg_assert(0);
@@ -101,12 +102,12 @@ void ML_(pp_TyEnt)( TyEnt* te )
          if (te->Te.Field.nLoc == -1)
             VG_(printf)("Te_Field(ty=0x%05lx,pos.offset=%ld,\"%s\")",
                         te->Te.Field.typeR, te->Te.Field.pos.offset,
-                        te->Te.Field.name ? te->Te.Field.name : (UChar*)"");
+                        te->Te.Field.name ? te->Te.Field.name : "");
          else
             VG_(printf)("Te_Field(ty=0x%05lx,nLoc=%lu,pos.loc=%p,\"%s\")",
                         te->Te.Field.typeR, te->Te.Field.nLoc,
                         te->Te.Field.pos.loc,
-                        te->Te.Field.name ? te->Te.Field.name : (UChar*)"");
+                        te->Te.Field.name ? te->Te.Field.name : "");
          break;
       case Te_Bound:
          VG_(printf)("Te_Bound[");
@@ -125,19 +126,29 @@ void ML_(pp_TyEnt)( TyEnt* te )
          VG_(printf)("Te_TyBase(%d,%c,\"%s\")",
                      te->Te.TyBase.szB, te->Te.TyBase.enc,
                      te->Te.TyBase.name ? te->Te.TyBase.name
-                                        : (UChar*)"(null)" );
+                                        : "(null)" );
          break;
-      case Te_TyPorR:
-         VG_(printf)("Te_TyPorR(%d,%c,0x%05lx)",
-                     te->Te.TyPorR.szB,
-                     te->Te.TyPorR.isPtr ? 'P' : 'R',
+      case Te_TyPtr:
+         VG_(printf)("Te_TyPtr(%d,0x%05lx)", te->Te.TyPorR.szB,
+                     te->Te.TyPorR.typeR);
+         break;
+      case Te_TyRef:
+         VG_(printf)("Te_TyRef(%d,0x%05lx)", te->Te.TyPorR.szB,
+                     te->Te.TyPorR.typeR);
+         break;
+      case Te_TyPtrMbr:
+         VG_(printf)("Te_TyMbr(%d,0x%05lx)", te->Te.TyPorR.szB,
+                     te->Te.TyPorR.typeR);
+         break;
+      case Te_TyRvalRef:
+         VG_(printf)("Te_TyRvalRef(%d,0x%05lx)", te->Te.TyPorR.szB,
                      te->Te.TyPorR.typeR);
          break;
       case Te_TyTyDef:
          VG_(printf)("Te_TyTyDef(0x%05lx,\"%s\")",
                      te->Te.TyTyDef.typeR,
                      te->Te.TyTyDef.name ? te->Te.TyTyDef.name
-                                         : (UChar*)"" );
+                                         : "" );
          break;
       case Te_TyStOrUn:
          if (te->Te.TyStOrUn.complete) {
@@ -146,7 +157,7 @@ void ML_(pp_TyEnt)( TyEnt* te )
                         te->Te.TyStOrUn.isStruct ? 'S' : 'U',
                         te->Te.TyStOrUn.fieldRs,
                         te->Te.TyStOrUn.name ? te->Te.TyStOrUn.name
-                                             : (UChar*)"" );
+                                             : "" );
             if (te->Te.TyStOrUn.fieldRs)
                pp_XArray_of_cuOffs( te->Te.TyStOrUn.fieldRs );
          } else {
@@ -158,7 +169,7 @@ void ML_(pp_TyEnt)( TyEnt* te )
          VG_(printf)("Te_TyEnum(%d,%p,\"%s\")",
                      te->Te.TyEnum.szB, te->Te.TyEnum.atomRs,
                      te->Te.TyEnum.name ? te->Te.TyEnum.name
-                                        : (UChar*)"" );
+                                        : "" );
          if (te->Te.TyEnum.atomRs)
             pp_XArray_of_cuOffs( te->Te.TyEnum.atomRs );
          break;
@@ -187,7 +198,7 @@ void ML_(pp_TyEnt)( TyEnt* te )
 
 /* Print a whole XArray of TyEnts, debug-style */
 
-void ML_(pp_TyEnts)( XArray* tyents, HChar* who )
+void ML_(pp_TyEnts)( XArray* tyents, const HChar* who )
 {
    Word i, n;
    VG_(printf)("------ %s ------\n", who);
@@ -237,19 +248,31 @@ void ML_(pp_TyEnt_C_ishly)( XArray* /* of TyEnt */ tyents,
          if (!ent->Te.TyBase.name) goto unhandled;
          VG_(printf)("%s", ent->Te.TyBase.name);
          break;
-      case Te_TyPorR:
+      case Te_TyPtr:
          ML_(pp_TyEnt_C_ishly)(tyents, ent->Te.TyPorR.typeR);
-         VG_(printf)("%s", ent->Te.TyPorR.isPtr ? "*" : "&");
+         VG_(printf)("*");
+         break;
+      case Te_TyRef:
+         ML_(pp_TyEnt_C_ishly)(tyents, ent->Te.TyPorR.typeR);
+         VG_(printf)("&");
+         break;
+      case Te_TyPtrMbr:
+         ML_(pp_TyEnt_C_ishly)(tyents, ent->Te.TyPorR.typeR);
+         VG_(printf)("*");
+         break;
+      case Te_TyRvalRef:
+         ML_(pp_TyEnt_C_ishly)(tyents, ent->Te.TyPorR.typeR);
+         VG_(printf)("&&");
          break;
       case Te_TyEnum:
-         if (!ent->Te.TyEnum.name) goto unhandled;
-         VG_(printf)("enum %s", ent->Te.TyEnum.name);
+         VG_(printf)("enum %s", ent->Te.TyEnum.name ? ent->Te.TyEnum.name
+                                                    : "<anonymous>" );
          break;
       case Te_TyStOrUn:
          VG_(printf)("%s %s",
                      ent->Te.TyStOrUn.isStruct ? "struct" : "union",
                      ent->Te.TyStOrUn.name ? ent->Te.TyStOrUn.name
-                                           : (UChar*)"<anonymous>" );
+                                           : "<anonymous>" );
          break;
       case Te_TyArray:
          ML_(pp_TyEnt_C_ishly)(tyents, ent->Te.TyArray.typeR);
@@ -264,8 +287,8 @@ void ML_(pp_TyEnt_C_ishly)( XArray* /* of TyEnt */ tyents,
          }
          break;
       case Te_TyTyDef:
-         if (!ent->Te.TyTyDef.name) goto unhandled;
-         VG_(printf)("%s", ent->Te.TyTyDef.name);
+         VG_(printf)("%s", ent->Te.TyTyDef.name ? ent->Te.TyTyDef.name
+                                                : "<anonymous>" );
          break;
       case Te_TyFn:
          VG_(printf)("%s", "<function_type>");
@@ -387,7 +410,7 @@ TyEnt* ML_(TyEnts__index_by_cuOff) ( XArray* /* of TyEnt */ ents,
 /* Generates a total ordering on TyEnts based only on their .cuOff
    fields. */
 
-Word ML_(TyEnt__cmp_by_cuOff_only) ( TyEnt* te1, TyEnt* te2 )
+Word ML_(TyEnt__cmp_by_cuOff_only) ( const TyEnt* te1, const TyEnt* te2 )
 {
    if (te1->cuOff < te2->cuOff) return -1;
    if (te1->cuOff > te2->cuOff) return 1;
@@ -397,29 +420,29 @@ Word ML_(TyEnt__cmp_by_cuOff_only) ( TyEnt* te1, TyEnt* te2 )
 
 /* Generates a total ordering on TyEnts based on everything except
    their .cuOff fields. */
-static __attribute__((always_inline)) Word UWord__cmp ( UWord a, UWord b ) {
+static inline Word UWord__cmp ( UWord a, UWord b ) {
    if (a < b) return -1;
    if (a > b) return 1;
    return 0;
 }
-static __attribute__((always_inline)) Word Long__cmp ( Long a, Long b ) {
+static inline Word Long__cmp ( Long a, Long b ) {
    if (a < b) return -1;
    if (a > b) return 1;
    return 0;
 }
-static __attribute__((always_inline)) Word Bool__cmp ( Bool a, Bool b ) {
+static inline Word Bool__cmp ( Bool a, Bool b ) {
    vg_assert( ((UWord)a) <= 1 );
    vg_assert( ((UWord)b) <= 1 );
    if (a < b) return -1;
    if (a > b) return 1;
    return 0;
 }
-static __attribute__((always_inline)) Word UChar__cmp ( UChar a, UChar b ) {
+static inline Word UChar__cmp ( UChar a, UChar b ) {
    if (a < b) return -1;
    if (a > b) return 1;
    return 0;
 }
-static __attribute__((always_inline)) Word Int__cmp ( Int a, Int b ) {
+static inline Word Int__cmp ( Int a, Int b ) {
    if (a < b) return -1;
    if (a > b) return 1;
    return 0;
@@ -446,7 +469,7 @@ static Word Bytevector__cmp ( UChar* a, UChar* b, Word n ) {
    }
    return 0;
 }
-static Word Asciiz__cmp ( UChar* a, UChar* b ) {
+static Word Asciiz__cmp ( const HChar* a, const HChar* b ) {
    /* A wrapper around strcmp that handles NULL strings safely. */
    if (a == NULL && b == NULL) return 0;
    if (a == NULL && b != NULL) return -1;
@@ -454,7 +477,7 @@ static Word Asciiz__cmp ( UChar* a, UChar* b ) {
    return VG_(strcmp)(a, b);
 }
 
-Word ML_(TyEnt__cmp_by_all_except_cuOff) ( TyEnt* te1, TyEnt* te2 )
+Word ML_(TyEnt__cmp_by_all_except_cuOff) ( const TyEnt* te1, const TyEnt* te2 )
 {
    Word r;
    if (te1->tag < te2->tag) return -1;
@@ -503,12 +526,13 @@ Word ML_(TyEnt__cmp_by_all_except_cuOff) ( TyEnt* te1, TyEnt* te2 )
       if (r != 0) return r;
       r = Asciiz__cmp(te1->Te.TyBase.name, te2->Te.TyBase.name);
       return r;
-   case Te_TyPorR:
+   case Te_TyPtr:
+   case Te_TyRef:
+   case Te_TyPtrMbr:
+   case Te_TyRvalRef:
       r = Int__cmp(te1->Te.TyPorR.szB, te2->Te.TyPorR.szB);
       if (r != 0) return r;
       r = UWord__cmp(te1->Te.TyPorR.typeR, te2->Te.TyPorR.typeR);
-      if (r != 0) return r;
-      r = Bool__cmp(te1->Te.TyPorR.isPtr, te2->Te.TyPorR.isPtr);
       return r;
    case Te_TyTyDef:
       r = UWord__cmp(te1->Te.TyTyDef.typeR, te2->Te.TyTyDef.typeR);
@@ -584,7 +608,10 @@ void ML_(TyEnt__make_EMPTY) ( TyEnt* te )
       case Te_TyBase:
          if (te->Te.TyBase.name) ML_(dinfo_free)(te->Te.TyBase.name);
          break;
-      case Te_TyPorR:
+      case Te_TyPtr:
+      case Te_TyRef:
+      case Te_TyPtrMbr:
+      case Te_TyRvalRef:
          break;
       case Te_TyTyDef:
          if (te->Te.TyTyDef.name) ML_(dinfo_free)(te->Te.TyTyDef.name);
@@ -661,7 +688,10 @@ MaybeULong ML_(sizeOfType)( XArray* /* of TyEnt */ tyents,
          if (ent2->tag == Te_UNKNOWN)
             return mk_MaybeULong_Nothing(); /*UNKNOWN*/
          return ML_(sizeOfType)( tyents, ent->Te.TyTyDef.typeR );
-      case Te_TyPorR:
+      case Te_TyPtr:
+      case Te_TyRef:
+      case Te_TyPtrMbr:
+      case Te_TyRvalRef:
          vg_assert(ent->Te.TyPorR.szB == 4 || ent->Te.TyPorR.szB == 8);
          return mk_MaybeULong_Just( ent->Te.TyPorR.szB );
       case Te_TyStOrUn:
@@ -706,15 +736,15 @@ MaybeULong ML_(sizeOfType)( XArray* /* of TyEnt */ tyents,
 /* Describe where in the type 'offset' falls.  Caller must
    deallocate the resulting XArray. */
 
-static void copy_UWord_into_XA ( XArray* /* of UChar */ xa,
+static void copy_UWord_into_XA ( XArray* /* of HChar */ xa,
                                  UWord uw ) {
-   UChar buf[32];
+   HChar buf[32];
    VG_(memset)(buf, 0, sizeof(buf));
    VG_(sprintf)(buf, "%lu", uw);
    VG_(addBytesToXA)( xa, buf, VG_(strlen)(buf));
 }
 
-XArray* /*UChar*/ ML_(describe_type)( /*OUT*/PtrdiffT* residual_offset,
+XArray* /*HChar*/ ML_(describe_type)( /*OUT*/PtrdiffT* residual_offset,
                                       XArray* /* of TyEnt */ tyents,
                                       UWord ty_cuOff, 
                                       PtrdiffT offset )
@@ -722,7 +752,7 @@ XArray* /*UChar*/ ML_(describe_type)( /*OUT*/PtrdiffT* residual_offset,
    TyEnt*  ty;
    XArray* xa = VG_(newXA)( ML_(dinfo_zalloc), "di.tytypes.dt.1",
                             ML_(dinfo_free),
-                            sizeof(UChar) );
+                            sizeof(HChar) );
    vg_assert(xa);
 
    ty = ML_(TyEnts__index_by_cuOff)(tyents, NULL, ty_cuOff);
@@ -738,7 +768,10 @@ XArray* /*UChar*/ ML_(describe_type)( /*OUT*/PtrdiffT* residual_offset,
          case Te_TyEnum:
          case Te_TyFn:
          case Te_TyVoid:
-         case Te_TyPorR:
+         case Te_TyPtr:
+         case Te_TyRef:
+         case Te_TyPtrMbr:
+         case Te_TyRvalRef:
          case Te_TyBase:
             goto done;
 
@@ -752,7 +785,8 @@ XArray* /*UChar*/ ML_(describe_type)( /*OUT*/PtrdiffT* residual_offset,
             PtrdiffT   offMin = 0, offMax1 = 0;
             if (!ty->Te.TyStOrUn.isStruct) goto done;
             fieldRs = ty->Te.TyStOrUn.fieldRs;
-            if ((!fieldRs) || VG_(sizeXA)(fieldRs) == 0) goto done;
+            if (((!fieldRs) || VG_(sizeXA)(fieldRs) == 0)
+                && (ty->Te.TyStOrUn.typeR == 0)) goto done;
             for (i = 0; i < VG_(sizeXA)( fieldRs ); i++ ) {
                fieldR = *(UWord*)VG_(indexXA)( fieldRs, i );
                field = ML_(TyEnts__index_by_cuOff)(tyents, NULL, fieldR);
@@ -798,8 +832,14 @@ XArray* /*UChar*/ ML_(describe_type)( /*OUT*/PtrdiffT* residual_offset,
             }
             /* Did we find a suitable field? */
             vg_assert(i >= 0 && i <= VG_(sizeXA)( fieldRs ));
-            if (i == VG_(sizeXA)( fieldRs ))
-               goto done; /* No.  Give up. */
+            if (i == VG_(sizeXA)( fieldRs )) {
+               ty = ML_(TyEnts__index_by_cuOff)(tyents, NULL,
+                                                   ty->Te.TyStOrUn.typeR);
+               vg_assert(ty);
+               if (ty->tag == Te_UNKNOWN) goto done;
+               vg_assert(ML_(TyEnt__is_type)(ty));
+               continue;
+            }
             /* Yes.  'field' is it. */
             vg_assert(field);
             if (!field->Te.Field.name) goto done;
