@@ -7,7 +7,7 @@
    This file is part of Valgrind, a dynamic binary instrumentation
    framework.
 
-   Copyright (C) 2005-2007 Nicholas Nethercote
+   Copyright (C) 2005-2013 Nicholas Nethercote
       njn@valgrind.org
 
    This program is free software; you can redistribute it and/or
@@ -47,7 +47,7 @@ struct _VgHashTable {
    UInt         iterChain;  // next chain to be traversed by the iterator
    VgHashNode** chains;     // expanding array of hash chains
    Bool         iterOK;     // table safe to iterate over?
-   HChar*       name;       // name of table (for debugging only)
+   const HChar* name;       // name of table (for debugging only)
 };
 
 #define N_HASH_PRIMES 20
@@ -64,13 +64,14 @@ static SizeT primes[N_HASH_PRIMES] = {
 /*--- Functions                                                    ---*/
 /*--------------------------------------------------------------------*/
 
-VgHashTable VG_(HT_construct) ( HChar* name )
+VgHashTable VG_(HT_construct) ( const HChar* name )
 {
    /* Initialises to zero, ie. all entries NULL */
    SizeT       n_chains = primes[0];
    SizeT       sz       = n_chains * sizeof(VgHashNode*);
-   VgHashTable table    = VG_(calloc)(1, sizeof(struct _VgHashTable));
-   table->chains        = VG_(calloc)(1, sz);
+   VgHashTable table    = VG_(calloc)("hashtable.Hc.1",
+                                      1, sizeof(struct _VgHashTable));
+   table->chains        = VG_(calloc)("hashtable.Hc.2", 1, sz);
    table->n_chains      = n_chains;
    table->n_elements    = 0;
    table->iterOK        = True;
@@ -119,7 +120,7 @@ static void resize ( VgHashTable table )
 
    table->n_chains = new_chains;
    sz = new_chains * sizeof(VgHashNode*);
-   chains = VG_(calloc)(1, sz);
+   chains = VG_(calloc)("hashtable.resize.1", 1, sz);
 
    for (i = 0; i < old_chains; i++) {
       node = table->chains[i];
@@ -137,7 +138,7 @@ static void resize ( VgHashTable table )
 }
 
 /* Puts a new, heap allocated VgHashNode, into the VgHashTable.  Prepends
-   the node to the appropriate chain. */
+   the node to the appropriate chain.  No duplicate key detection is done. */
 void VG_(HT_add_node) ( VgHashTable table, void* vnode )
 {
    VgHashNode* node     = (VgHashNode*)vnode;
@@ -189,10 +190,9 @@ void* VG_(HT_remove) ( VgHashTable table, UWord key )
    return NULL;
 }
 
-/* Allocates a suitably-sized array, copies all the hashtable elements
-   into it, then returns both the array and the size of it.  This is
-   used by the memory-leak detector.  The array must be freed with
-   VG_(free).
+/* Allocates a suitably-sized array, copies pointers to all the hashtable
+   elements into it, then returns both the array and the size of it.  The
+   array must be freed with VG_(free).
 */
 VgHashNode** VG_(HT_to_array) ( VgHashTable table, /*OUT*/ UInt* n_elems )
 {
@@ -200,16 +200,11 @@ VgHashNode** VG_(HT_to_array) ( VgHashTable table, /*OUT*/ UInt* n_elems )
    VgHashNode** arr;
    VgHashNode*  node;
 
-   *n_elems = 0;
-   for (i = 0; i < table->n_chains; i++) {
-      for (node = table->chains[i]; node != NULL; node = node->next) {
-         (*n_elems)++;
-      }
-   }
+   *n_elems = table->n_elements;
    if (*n_elems == 0)
       return NULL;
 
-   arr = VG_(malloc)( *n_elems * sizeof(VgHashNode*) );
+   arr = VG_(malloc)( "hashtable.Hta.1", *n_elems * sizeof(VgHashNode*) );
 
    j = 0;
    for (i = 0; i < table->n_chains; i++) {
@@ -254,7 +249,7 @@ void* VG_(HT_Next)(VgHashTable table)
    return NULL;
 }
 
-void VG_(HT_destruct)(VgHashTable table)
+void VG_(HT_destruct)(VgHashTable table, void(*freenode_fn)(void*))
 {
    UInt       i;
    VgHashNode *node, *node_next;
@@ -262,7 +257,7 @@ void VG_(HT_destruct)(VgHashTable table)
    for (i = 0; i < table->n_chains; i++) {
       for (node = table->chains[i]; node != NULL; node = node_next) {
          node_next = node->next;
-         VG_(free)(node);
+         freenode_fn(node);
       }
    }
    VG_(free)(table->chains);
