@@ -8,9 +8,9 @@
    This file is part of Valgrind, a dynamic binary instrumentation
    framework.
 
-   Copyright (C) 2000-2008 Nicholas Nethercote
+   Copyright (C) 2000-2013 Nicholas Nethercote
       njn@valgrind.org
-   Copyright (C) 2004-2008 Paul Mackerras
+   Copyright (C) 2004-2013 Paul Mackerras
       paulus@samba.org
 
    This program is free software; you can redistribute it and/or
@@ -31,9 +31,12 @@
    The GNU General Public License is contained in the file COPYING.
 */
 
+#if defined(VGP_ppc64_linux)
+
 #include "pub_core_basics.h"
 #include "pub_core_vki.h"
 #include "pub_core_vkiscnums.h"
+#include "pub_core_libcsetjmp.h"    // to keep _threadstate.h happy
 #include "pub_core_threadstate.h"
 #include "pub_core_aspacemgr.h"
 #include "pub_core_libcbase.h"
@@ -148,12 +151,12 @@ static Bool extend ( ThreadState *tst, Addr addr, SizeT size )
    if (stackseg == NULL || !stackseg->hasR || !stackseg->hasW) {
       VG_(message)(
          Vg_UserMsg,
-         "Can't extend stack to %#lx during signal delivery for thread %d:",
+         "Can't extend stack to %#lx during signal delivery for thread %d:\n",
          addr, tid);
       if (stackseg == NULL)
-         VG_(message)(Vg_UserMsg, "  no stack segment");
+         VG_(message)(Vg_UserMsg, "  no stack segment\n");
       else
-         VG_(message)(Vg_UserMsg, "  too small or bad protection modes");
+         VG_(message)(Vg_UserMsg, "  too small or bad protection modes\n");
 
       /* set SIGSEGV to default handler */
       VG_(set_default_handler)(VKI_SIGSEGV);
@@ -187,7 +190,7 @@ void VG_(sigframe_create)( ThreadId tid,
    Addr sp;
    ThreadState* tst;
    Int sigNo = siginfo->si_signo;
-   Addr faultaddr;
+   /* Addr faultaddr; */ /* UNUSED */
    struct rt_sigframe* frame;
 
    /* Stack must be 16-byte aligned */
@@ -223,9 +226,11 @@ void VG_(sigframe_create)( ThreadId tid,
    VG_TRACK( post_mem_write, Vg_CoreSignal, tid, 
              sp, sizeof(UWord) );
 
+   /* UNUSED:
    faultaddr = (Addr)siginfo->_sifields._sigfault._addr;
    if (sigNo == VKI_SIGILL && siginfo->si_code > 0)
       faultaddr = tst->arch.vex.guest_CIA;
+   */
 
    VG_(memcpy)(&frame->info, siginfo, sizeof(*siginfo));
    VG_TRACK( post_mem_write, Vg_CoreSignal, tid,
@@ -375,13 +380,16 @@ void VG_(sigframe_destroy)( ThreadId tid, Bool isRT )
 
    if (VG_(clo_trace_signals))
       VG_(message)(Vg_DebugMsg,
-                   "vg_pop_signal_frame (thread %d): isRT=%d valid magic; EIP=%#llx",
+                   "vg_pop_signal_frame (thread %d): isRT=%d "
+                   "valid magic; EIP=%#llx\n",
                    tid, has_siginfo, tst->arch.vex.guest_CIA);
 
    /* tell the tools */
    VG_TRACK( post_deliver_signal, tid, sigNo );
 }
 
+#endif // defined(VGP_ppc64_linux)
+
 /*--------------------------------------------------------------------*/
-/*--- end                                   sigframe-ppc64-linux.c ---*/
+/*--- end                                                          ---*/
 /*--------------------------------------------------------------------*/
