@@ -7,7 +7,7 @@
    This file is part of Valgrind, a dynamic binary instrumentation
    framework.
 
-   Copyright (C) 2000-2009 Nicholas Nethercote
+   Copyright (C) 2000-2013 Nicholas Nethercote
       njn@valgrind.org
 
    This program is free software; you can redistribute it and/or
@@ -31,7 +31,8 @@
 #ifndef __PRIV_SYSWRAP_LINUX_H
 #define __PRIV_SYSWRAP_LINUX_H
 
-/* requires #include "priv_types_n_macros.h" */
+#include "pub_core_basics.h"     // ThreadId
+#include "priv_types_n_macros.h" // DECL_TEMPLATE
 
 // Clone-related functions
 extern Word ML_(start_thread_NORETURN) ( void* arg );
@@ -47,6 +48,18 @@ extern SysRes ML_(do_fork_clone) ( ThreadId tid, UInt flags,
 DECL_TEMPLATE(linux, sys_mount);
 DECL_TEMPLATE(linux, sys_oldumount);
 DECL_TEMPLATE(linux, sys_umount);
+DECL_TEMPLATE(linux, sys_perf_event_open);
+DECL_TEMPLATE(linux, sys_preadv);
+DECL_TEMPLATE(linux, sys_pwritev);
+DECL_TEMPLATE(linux, sys_sendmmsg);
+DECL_TEMPLATE(linux, sys_recvmmsg);
+DECL_TEMPLATE(linux, sys_dup3);
+DECL_TEMPLATE(linux, sys_getcpu);
+DECL_TEMPLATE(linux, sys_splice);
+DECL_TEMPLATE(linux, sys_tee);
+DECL_TEMPLATE(linux, sys_vmsplice);
+DECL_TEMPLATE(linux, sys_readahead);
+DECL_TEMPLATE(linux, sys_move_pages);
 
 // POSIX, but various sub-cases differ between Linux and Darwin.
 DECL_TEMPLATE(linux, sys_fcntl);
@@ -94,6 +107,8 @@ DECL_TEMPLATE(linux, sys_eventfd2);
 
 DECL_TEMPLATE(linux, sys_fallocate);
 
+DECL_TEMPLATE(linux, sys_prlimit64);
+
 DECL_TEMPLATE(linux, sys_gettid);
 DECL_TEMPLATE(linux, sys_set_tid_address);
 DECL_TEMPLATE(linux, sys_tkill);
@@ -116,6 +131,7 @@ DECL_TEMPLATE(linux, sys_set_mempolicy);
 DECL_TEMPLATE(linux, sys_get_mempolicy);
 
 DECL_TEMPLATE(linux, sys_inotify_init);
+DECL_TEMPLATE(linux, sys_inotify_init1);
 DECL_TEMPLATE(linux, sys_inotify_add_watch);
 DECL_TEMPLATE(linux, sys_inotify_rm_watch);
 
@@ -130,6 +146,7 @@ DECL_TEMPLATE(linux, sys_clock_settime);
 DECL_TEMPLATE(linux, sys_clock_gettime);
 DECL_TEMPLATE(linux, sys_clock_getres);
 DECL_TEMPLATE(linux, sys_clock_nanosleep);
+DECL_TEMPLATE(linux, sys_clock_adjtime);
 
 DECL_TEMPLATE(linux, sys_timer_create);      // Linux: varies across archs?
 DECL_TEMPLATE(linux, sys_timer_settime);
@@ -160,6 +177,8 @@ DECL_TEMPLATE(linux, sys_readlinkat);
 DECL_TEMPLATE(linux, sys_fchmodat);
 DECL_TEMPLATE(linux, sys_faccessat);
 DECL_TEMPLATE(linux, sys_utimensat);
+DECL_TEMPLATE(linux, sys_name_to_handle_at);
+DECL_TEMPLATE(linux, sys_open_by_handle_at);
 
 DECL_TEMPLATE(linux, sys_add_key);
 DECL_TEMPLATE(linux, sys_request_key);
@@ -209,7 +228,7 @@ DECL_TEMPLATE(linux, sys_sched_getscheduler);
 DECL_TEMPLATE(linux, sys_sched_yield);
 DECL_TEMPLATE(linux, sys_sched_get_priority_max);
 DECL_TEMPLATE(linux, sys_sched_get_priority_min);
-//DECL_TEMPLATE(linux, sys_sched_rr_get_interval);    // not yet encountered
+DECL_TEMPLATE(linux, sys_sched_rr_get_interval);
 DECL_TEMPLATE(linux, sys_sched_setaffinity);
 DECL_TEMPLATE(linux, sys_sched_getaffinity);
 
@@ -230,6 +249,7 @@ DECL_TEMPLATE(linux, sys_lseek);
 // Darwin (and probably other OSes) don't have the old_sigset_t type.
 DECL_TEMPLATE(linux, sys_sigpending);
 DECL_TEMPLATE(linux, sys_sigprocmask);
+DECL_TEMPLATE(linux, sys_sigaction);
 
 // I think these are Linux-specific?
 DECL_TEMPLATE(linux, sys_rt_sigaction);
@@ -237,10 +257,12 @@ DECL_TEMPLATE(linux, sys_rt_sigprocmask);
 DECL_TEMPLATE(linux, sys_rt_sigpending);
 DECL_TEMPLATE(linux, sys_rt_sigtimedwait);
 DECL_TEMPLATE(linux, sys_rt_sigqueueinfo);
+DECL_TEMPLATE(linux, sys_rt_tgsigqueueinfo);
 DECL_TEMPLATE(linux, sys_rt_sigsuspend);
 
 // Linux-specific?
 DECL_TEMPLATE(linux, sys_sync_file_range);
+DECL_TEMPLATE(linux, sys_sync_file_range2);
 DECL_TEMPLATE(linux, sys_stime);  /* maybe generic?  I'm not sure */
 
 // Linux specific (kernel modules)
@@ -250,6 +272,14 @@ DECL_TEMPLATE(linux, sys_delete_module);
 // Linux-specific (oprofile-related)
 DECL_TEMPLATE(linux, sys_lookup_dcookie);        // (*/32/64) L
 
+// Linux-specific (new in Linux 3.2)
+DECL_TEMPLATE(linux, sys_process_vm_readv);
+DECL_TEMPLATE(linux, sys_process_vm_writev);
+
+// Linux-specific (new in Linux 2.6.36)
+DECL_TEMPLATE(linux, sys_fanotify_init);
+DECL_TEMPLATE(linux, sys_fanotify_mark);
+
 /* ---------------------------------------------------------------------
    Wrappers for sockets and ipc-ery.  These are split into standalone
    procedures because x86-linux hides them inside multiplexors
@@ -267,11 +297,63 @@ extern void ML_(linux_PRE_sys_msgctl)      ( TId, UW, UW, UW );
 extern void ML_(linux_POST_sys_msgctl)     ( TId, UW, UW, UW, UW );
 extern void ML_(linux_PRE_sys_getsockopt)  ( TId, UW, UW, UW, UW, UW );
 extern void ML_(linux_POST_sys_getsockopt) ( TId, SR, UW, UW, UW, UW, UW );
+extern void ML_(linux_PRE_sys_setsockopt)  ( TId, UW, UW, UW, UW, UW );
+
+// Linux-specific (but non-arch-specific) ptrace wrapper helpers
+extern void ML_(linux_PRE_getregset) ( ThreadId, long, long );
+extern void ML_(linux_PRE_setregset) ( ThreadId, long, long );
+extern void ML_(linux_POST_getregset)( ThreadId, long, long );
 
 #undef TId
 #undef UW
 #undef SR
 
+/* sys_ipc and sys_socketcall are multiplexors which implements several syscalls.
+   Used e.g. by x86, ppc32, ppc64, ... */
+DECL_TEMPLATE(linux, sys_ipc);
+DECL_TEMPLATE(linux, sys_socketcall);
+
+/* Depending on the platform, the below are implemented as
+   direct syscalls or via the above sys_socketcall multiplexor. */
+
+/* Direct ipc related syscalls. */
+/* Semaphore */
+DECL_TEMPLATE(linux, sys_semget);
+DECL_TEMPLATE(linux, sys_semop);
+DECL_TEMPLATE(linux, sys_semctl);
+DECL_TEMPLATE(linux, sys_semtimedop);
+/* Shared memory */
+DECL_TEMPLATE(linux, wrap_sys_shmat);
+DECL_TEMPLATE(linux, sys_shmget);
+DECL_TEMPLATE(linux, sys_shmdt);
+DECL_TEMPLATE(linux, sys_shmctl);
+/* Message queue */
+DECL_TEMPLATE(linux, sys_msgget);
+DECL_TEMPLATE(linux, sys_msgrcv);
+DECL_TEMPLATE(linux, sys_msgsnd);
+DECL_TEMPLATE(linux, sys_msgctl);
+
+/* Direct socket related syscalls. */
+DECL_TEMPLATE(linux, sys_socket);
+DECL_TEMPLATE(linux, sys_setsockopt);
+DECL_TEMPLATE(linux, sys_getsockopt);
+DECL_TEMPLATE(linux, sys_connect);
+DECL_TEMPLATE(linux, sys_accept);
+DECL_TEMPLATE(linux, sys_accept4);
+DECL_TEMPLATE(linux, sys_send);
+DECL_TEMPLATE(linux, sys_sendto);
+DECL_TEMPLATE(linux, sys_recv);
+DECL_TEMPLATE(linux, sys_recvfrom);
+DECL_TEMPLATE(linux, sys_sendmsg);
+DECL_TEMPLATE(linux, sys_recvmsg);
+DECL_TEMPLATE(linux, sys_shutdown);
+DECL_TEMPLATE(linux, sys_bind);
+DECL_TEMPLATE(linux, sys_listen);
+DECL_TEMPLATE(linux, sys_getsockname);
+DECL_TEMPLATE(linux, sys_getpeername);
+DECL_TEMPLATE(linux, sys_socketpair);
+
+
 #endif   // __PRIV_SYSWRAP_LINUX_H
 
 /*--------------------------------------------------------------------*/
