@@ -7,7 +7,7 @@
    This file is part of Valgrind, a dynamic binary instrumentation
    framework.
 
-   Copyright (C) 2000-2010 Nicholas Nethercote
+   Copyright (C) 2000-2013 Nicholas Nethercote
       njn@valgrind.org
 
    This program is free software; you can redistribute it and/or
@@ -33,6 +33,7 @@
 #include "pub_core_basics.h"
 #include "pub_core_vki.h"
 #include "pub_core_vkiscnums.h"
+#include "pub_core_libcsetjmp.h"    // to keep _threadstate.h happy
 #include "pub_core_threadstate.h"
 #include "pub_core_aspacemgr.h"
 #include "pub_core_debuglog.h"
@@ -141,6 +142,7 @@ Long do_syscall_clone_amd64_linux ( Word (*fn)(void *),
                                     vki_modify_ldt_t * );
 asm(
 ".text\n"
+".globl do_syscall_clone_amd64_linux\n"
 "do_syscall_clone_amd64_linux:\n"
         // set up child stack, temporarily preserving fn and arg
 "       subq    $16, %rsi\n"            // make space on stack
@@ -288,6 +290,7 @@ static SysRes do_clone ( ThreadId ptid,
       know that this thread has come into existence.  If the clone
       fails, we'll send out a ll_exit notification for it at the out:
       label below, to clean up. */
+   vg_assert(VG_(owns_BigLock_LL)(ptid));
    VG_TRACK ( pre_thread_ll_create, ptid, ctid );
 
    if (flags & VKI_CLONE_SETTLS) {
@@ -355,35 +358,6 @@ void setup_child ( /*OUT*/ ThreadArchState *child,
    magic. */
 DECL_TEMPLATE(amd64_linux, sys_clone);
 DECL_TEMPLATE(amd64_linux, sys_rt_sigreturn);
-DECL_TEMPLATE(amd64_linux, sys_socket);
-DECL_TEMPLATE(amd64_linux, sys_setsockopt);
-DECL_TEMPLATE(amd64_linux, sys_getsockopt);
-DECL_TEMPLATE(amd64_linux, sys_connect);
-DECL_TEMPLATE(amd64_linux, sys_accept);
-DECL_TEMPLATE(amd64_linux, sys_accept4);
-DECL_TEMPLATE(amd64_linux, sys_sendto);
-DECL_TEMPLATE(amd64_linux, sys_recvfrom);
-DECL_TEMPLATE(amd64_linux, sys_sendmsg);
-DECL_TEMPLATE(amd64_linux, sys_recvmsg);
-DECL_TEMPLATE(amd64_linux, sys_shutdown);
-DECL_TEMPLATE(amd64_linux, sys_bind);
-DECL_TEMPLATE(amd64_linux, sys_listen);
-DECL_TEMPLATE(amd64_linux, sys_getsockname);
-DECL_TEMPLATE(amd64_linux, sys_getpeername);
-DECL_TEMPLATE(amd64_linux, sys_socketpair);
-DECL_TEMPLATE(amd64_linux, sys_semget);
-DECL_TEMPLATE(amd64_linux, sys_semop);
-DECL_TEMPLATE(amd64_linux, sys_semtimedop);
-DECL_TEMPLATE(amd64_linux, sys_semctl);
-DECL_TEMPLATE(amd64_linux, sys_msgget);
-DECL_TEMPLATE(amd64_linux, sys_msgrcv);
-DECL_TEMPLATE(amd64_linux, sys_msgsnd);
-DECL_TEMPLATE(amd64_linux, sys_msgctl);
-DECL_TEMPLATE(amd64_linux, sys_shmget);
-DECL_TEMPLATE(amd64_linux, wrap_sys_shmat);
-DECL_TEMPLATE(amd64_linux, sys_shmdt);
-DECL_TEMPLATE(amd64_linux, sys_shmdt);
-DECL_TEMPLATE(amd64_linux, sys_shmctl);
 DECL_TEMPLATE(amd64_linux, sys_arch_prctl);
 DECL_TEMPLATE(amd64_linux, sys_ptrace);
 DECL_TEMPLATE(amd64_linux, sys_fadvise64);
@@ -396,21 +370,35 @@ PRE(sys_clone)
    ULong cloneflags;
 
    PRINT("sys_clone ( %lx, %#lx, %#lx, %#lx, %#lx )",ARG1,ARG2,ARG3,ARG4,ARG5);
-   PRE_REG_READ5(int, "clone",
+   PRE_REG_READ2(int, "clone",
                  unsigned long, flags,
-                 void *, child_stack,
-                 int *, parent_tidptr,
-                 int *, child_tidptr,
-                 void *, tlsaddr);
+                 void *, child_stack);
 
    if (ARG1 & VKI_CLONE_PARENT_SETTID) {
+      if (VG_(tdict).track_pre_reg_read) {
+         PRA3("clone", int *, parent_tidptr);
+      }
       PRE_MEM_WRITE("clone(parent_tidptr)", ARG3, sizeof(Int));
       if (!VG_(am_is_valid_for_client)(ARG3, sizeof(Int), VKI_PROT_WRITE)) {
          SET_STATUS_Failure( VKI_EFAULT );
          return;
       }
    }
+   if (ARG1 & VKI_CLONE_SETTLS) {
+      if (VG_(tdict).track_pre_reg_read) {
+         PRA4("clone", vki_modify_ldt_t *, tlsinfo);
+      }
+      PRE_MEM_READ("clone(tlsinfo)", ARG4, sizeof(vki_modify_ldt_t));
+      if (!VG_(am_is_valid_for_client)(ARG4, sizeof(vki_modify_ldt_t), 
+                                             VKI_PROT_READ)) {
+         SET_STATUS_Failure( VKI_EFAULT );
+         return;
+      }
+   }
    if (ARG1 & (VKI_CLONE_CHILD_SETTID | VKI_CLONE_CHILD_CLEARTID)) {
+      if (VG_(tdict).track_pre_reg_read) {
+         PRA5("clone", int *, child_tidptr);
+      }
       PRE_MEM_WRITE("clone(child_tidptr)", ARG4, sizeof(Int));
       if (!VG_(am_is_valid_for_client)(ARG4, sizeof(Int), VKI_PROT_WRITE)) {
          SET_STATUS_Failure( VKI_EFAULT );
@@ -600,6 +588,12 @@ PRE(sys_ptrace)
    case VKI_PTRACE_SETSIGINFO:
       PRE_MEM_READ( "ptrace(setsiginfo)", ARG4, sizeof(vki_siginfo_t));
       break;
+   case VKI_PTRACE_GETREGSET:
+      ML_(linux_PRE_getregset)(tid, ARG3, ARG4);
+      break;
+   case VKI_PTRACE_SETREGSET:
+      ML_(linux_PRE_setregset)(tid, ARG3, ARG4);
+      break;
    default:
       break;
    }
@@ -628,352 +622,14 @@ POST(sys_ptrace)
        */
       POST_MEM_WRITE( ARG4, sizeof(vki_siginfo_t));
       break;
+   case VKI_PTRACE_GETREGSET:
+      ML_(linux_POST_getregset)(tid, ARG3, ARG4);
+      break;
    default:
       break;
    }
 }
 
-PRE(sys_socket)
-{
-   PRINT("sys_socket ( %ld, %ld, %ld )",ARG1,ARG2,ARG3);
-   PRE_REG_READ3(long, "socket", int, domain, int, type, int, protocol);
-}
-POST(sys_socket)
-{
-   SysRes r;
-   vg_assert(SUCCESS);
-   r = ML_(generic_POST_sys_socket)(tid, VG_(mk_SysRes_Success)(RES));
-   SET_STATUS_from_SysRes(r);
-}
-
-PRE(sys_setsockopt)
-{
-   PRINT("sys_setsockopt ( %ld, %ld, %ld, %#lx, %ld )",ARG1,ARG2,ARG3,ARG4,ARG5);
-   PRE_REG_READ5(long, "setsockopt",
-                 int, s, int, level, int, optname,
-                 const void *, optval, int, optlen);
-   ML_(generic_PRE_sys_setsockopt)(tid, ARG1,ARG2,ARG3,ARG4,ARG5);
-}
-
-PRE(sys_getsockopt)
-{
-   PRINT("sys_getsockopt ( %ld, %ld, %ld, %#lx, %#lx )",ARG1,ARG2,ARG3,ARG4,ARG5);
-   PRE_REG_READ5(long, "getsockopt",
-                 int, s, int, level, int, optname,
-                 void *, optval, int, *optlen);
-   ML_(linux_PRE_sys_getsockopt)(tid, ARG1,ARG2,ARG3,ARG4,ARG5);
-}
-POST(sys_getsockopt)
-{
-   vg_assert(SUCCESS);
-   ML_(linux_POST_sys_getsockopt)(tid, VG_(mk_SysRes_Success)(RES),
-                                       ARG1,ARG2,ARG3,ARG4,ARG5);
-}
-
-PRE(sys_connect)
-{
-   *flags |= SfMayBlock;
-   PRINT("sys_connect ( %ld, %#lx, %ld )",ARG1,ARG2,ARG3);
-   PRE_REG_READ3(long, "connect",
-                 int, sockfd, struct sockaddr *, serv_addr, int, addrlen);
-   ML_(generic_PRE_sys_connect)(tid, ARG1,ARG2,ARG3);
-}
-
-PRE(sys_accept)
-{
-   *flags |= SfMayBlock;
-   PRINT("sys_accept ( %ld, %#lx, %ld )",ARG1,ARG2,ARG3);
-   PRE_REG_READ3(long, "accept",
-                 int, s, struct sockaddr *, addr, int, *addrlen);
-   ML_(generic_PRE_sys_accept)(tid, ARG1,ARG2,ARG3);
-}
-POST(sys_accept)
-{
-   SysRes r;
-   vg_assert(SUCCESS);
-   r = ML_(generic_POST_sys_accept)(tid, VG_(mk_SysRes_Success)(RES),
-                                         ARG1,ARG2,ARG3);
-   SET_STATUS_from_SysRes(r);
-}
-
-PRE(sys_accept4)
-{
-   *flags |= SfMayBlock;
-   PRINT("sys_accept4 ( %ld, %#lx, %ld, %ld )",ARG1,ARG2,ARG3,ARG4);
-   PRE_REG_READ4(long, "accept4",
-                 int, s, struct sockaddr *, addr, int, *addrlen, int, flags);
-   ML_(generic_PRE_sys_accept)(tid, ARG1,ARG2,ARG3);
-}
-POST(sys_accept4)
-{
-   SysRes r;
-   vg_assert(SUCCESS);
-   r = ML_(generic_POST_sys_accept)(tid, VG_(mk_SysRes_Success)(RES),
-                                         ARG1,ARG2,ARG3);
-   SET_STATUS_from_SysRes(r);
-}
-
-PRE(sys_sendto)
-{
-   *flags |= SfMayBlock;
-   PRINT("sys_sendto ( %ld, %#lx, %ld, %lu, %#lx, %ld )",ARG1,ARG2,ARG3,ARG4,ARG5,ARG6);
-   PRE_REG_READ6(long, "sendto",
-                 int, s, const void *, msg, int, len, 
-                 unsigned int, flags, 
-                 const struct sockaddr *, to, int, tolen);
-   ML_(generic_PRE_sys_sendto)(tid, ARG1,ARG2,ARG3,ARG4,ARG5,ARG6);
-}
-
-PRE(sys_recvfrom)
-{
-   *flags |= SfMayBlock;
-   PRINT("sys_recvfrom ( %ld, %#lx, %ld, %lu, %#lx, %#lx )",ARG1,ARG2,ARG3,ARG4,ARG5,ARG6);
-   PRE_REG_READ6(long, "recvfrom",
-                 int, s, void *, buf, int, len, unsigned int, flags,
-                 struct sockaddr *, from, int *, fromlen);
-   ML_(generic_PRE_sys_recvfrom)(tid, ARG1,ARG2,ARG3,ARG4,ARG5,ARG6);
-}
-POST(sys_recvfrom)
-{
-   vg_assert(SUCCESS);
-   ML_(generic_POST_sys_recvfrom)(tid, VG_(mk_SysRes_Success)(RES),
-                                       ARG1,ARG2,ARG3,ARG4,ARG5,ARG6);
-}
-
-PRE(sys_sendmsg)
-{
-   *flags |= SfMayBlock;
-   PRINT("sys_sendmsg ( %ld, %#lx, %ld )",ARG1,ARG2,ARG3);
-   PRE_REG_READ3(long, "sendmsg",
-                 int, s, const struct msghdr *, msg, int, flags);
-   ML_(generic_PRE_sys_sendmsg)(tid, ARG1,ARG2);
-}
-
-PRE(sys_recvmsg)
-{
-   *flags |= SfMayBlock;
-   PRINT("sys_recvmsg ( %ld, %#lx, %ld )",ARG1,ARG2,ARG3);
-   PRE_REG_READ3(long, "recvmsg", int, s, struct msghdr *, msg, int, flags);
-   ML_(generic_PRE_sys_recvmsg)(tid, ARG1,ARG2);
-}
-POST(sys_recvmsg)
-{
-   ML_(generic_POST_sys_recvmsg)(tid, ARG1,ARG2);
-}
-
-PRE(sys_shutdown)
-{
-   *flags |= SfMayBlock;
-   PRINT("sys_shutdown ( %ld, %ld )",ARG1,ARG2);
-   PRE_REG_READ2(int, "shutdown", int, s, int, how);
-}
-
-PRE(sys_bind)
-{
-   PRINT("sys_bind ( %ld, %#lx, %ld )",ARG1,ARG2,ARG3);
-   PRE_REG_READ3(long, "bind",
-                 int, sockfd, struct sockaddr *, my_addr, int, addrlen);
-   ML_(generic_PRE_sys_bind)(tid, ARG1,ARG2,ARG3);
-}
-
-PRE(sys_listen)
-{
-   PRINT("sys_listen ( %ld, %ld )",ARG1,ARG2);
-   PRE_REG_READ2(long, "listen", int, s, int, backlog);
-}
-
-PRE(sys_getsockname)
-{
-   PRINT("sys_getsockname ( %ld, %#lx, %#lx )",ARG1,ARG2,ARG3);
-   PRE_REG_READ3(long, "getsockname",
-                 int, s, struct sockaddr *, name, int *, namelen);
-   ML_(generic_PRE_sys_getsockname)(tid, ARG1,ARG2,ARG3);
-}
-POST(sys_getsockname)
-{
-   vg_assert(SUCCESS);
-   ML_(generic_POST_sys_getsockname)(tid, VG_(mk_SysRes_Success)(RES),
-                                          ARG1,ARG2,ARG3);
-}
-
-PRE(sys_getpeername)
-{
-   PRINT("sys_getpeername ( %ld, %#lx, %#lx )",ARG1,ARG2,ARG3);
-   PRE_REG_READ3(long, "getpeername",
-                 int, s, struct sockaddr *, name, int *, namelen);
-   ML_(generic_PRE_sys_getpeername)(tid, ARG1,ARG2,ARG3);
-}
-POST(sys_getpeername)
-{
-   vg_assert(SUCCESS);
-   ML_(generic_POST_sys_getpeername)(tid, VG_(mk_SysRes_Success)(RES),
-                                          ARG1,ARG2,ARG3);
-}
-
-PRE(sys_socketpair)
-{
-   PRINT("sys_socketpair ( %ld, %ld, %ld, %#lx )",ARG1,ARG2,ARG3,ARG4);
-   PRE_REG_READ4(long, "socketpair",
-                 int, d, int, type, int, protocol, int*, sv);
-   ML_(generic_PRE_sys_socketpair)(tid, ARG1,ARG2,ARG3,ARG4);
-}
-POST(sys_socketpair)
-{
-   vg_assert(SUCCESS);
-   ML_(generic_POST_sys_socketpair)(tid, VG_(mk_SysRes_Success)(RES),
-                                         ARG1,ARG2,ARG3,ARG4);
-}
-
-PRE(sys_semget)
-{
-   PRINT("sys_semget ( %ld, %ld, %ld )",ARG1,ARG2,ARG3);
-   PRE_REG_READ3(long, "semget", vki_key_t, key, int, nsems, int, semflg);
-}
-
-PRE(sys_semop)
-{
-   *flags |= SfMayBlock;
-   PRINT("sys_semop ( %ld, %#lx, %lu )",ARG1,ARG2,ARG3);
-   PRE_REG_READ3(long, "semop",
-                 int, semid, struct sembuf *, sops, unsigned, nsoops);
-   ML_(generic_PRE_sys_semop)(tid, ARG1,ARG2,ARG3);
-}
-
-PRE(sys_semtimedop)
-{
-   *flags |= SfMayBlock;
-   PRINT("sys_semtimedop ( %ld, %#lx, %lu, %#lx )",ARG1,ARG2,ARG3,ARG4);
-   PRE_REG_READ4(long, "semtimedop",
-                 int, semid, struct sembuf *, sops, unsigned, nsoops,
-                 struct timespec *, timeout);
-   ML_(generic_PRE_sys_semtimedop)(tid, ARG1,ARG2,ARG3,ARG4);
-}
-
-PRE(sys_semctl)
-{
-   switch (ARG3 & ~VKI_IPC_64) {
-   case VKI_IPC_INFO:
-   case VKI_SEM_INFO:
-      PRINT("sys_semctl ( %ld, %ld, %ld, %#lx )",ARG1,ARG2,ARG3,ARG4);
-      PRE_REG_READ4(long, "semctl",
-                    int, semid, int, semnum, int, cmd, struct seminfo *, arg);
-      break;
-   case VKI_IPC_STAT:
-   case VKI_SEM_STAT:
-   case VKI_IPC_SET:
-      PRINT("sys_semctl ( %ld, %ld, %ld, %#lx )",ARG1,ARG2,ARG3,ARG4);
-      PRE_REG_READ4(long, "semctl",
-                    int, semid, int, semnum, int, cmd, struct semid_ds *, arg);
-      break;
-   case VKI_GETALL:
-   case VKI_SETALL:
-      PRINT("sys_semctl ( %ld, %ld, %ld, %#lx )",ARG1,ARG2,ARG3,ARG4);
-      PRE_REG_READ4(long, "semctl",
-                    int, semid, int, semnum, int, cmd, unsigned short *, arg);
-      break;
-   default:
-      PRINT("sys_semctl ( %ld, %ld, %ld )",ARG1,ARG2,ARG3);
-      PRE_REG_READ3(long, "semctl",
-                    int, semid, int, semnum, int, cmd);
-      break;
-   }
-   ML_(generic_PRE_sys_semctl)(tid, ARG1,ARG2,ARG3|VKI_IPC_64,ARG4);
-}
-POST(sys_semctl)
-{
-   ML_(generic_POST_sys_semctl)(tid, RES,ARG1,ARG2,ARG3|VKI_IPC_64,ARG4);
-}
-
-PRE(sys_msgget)
-{
-   PRINT("sys_msgget ( %ld, %ld )",ARG1,ARG2);
-   PRE_REG_READ2(long, "msgget", vki_key_t, key, int, msgflg);
-}
-
-PRE(sys_msgsnd)
-{
-   PRINT("sys_msgsnd ( %ld, %#lx, %ld, %ld )",ARG1,ARG2,ARG3,ARG4);
-   PRE_REG_READ4(long, "msgsnd",
-                 int, msqid, struct msgbuf *, msgp, vki_size_t, msgsz, int, msgflg);
-   ML_(linux_PRE_sys_msgsnd)(tid, ARG1,ARG2,ARG3,ARG4);
-   if ((ARG4 & VKI_IPC_NOWAIT) == 0)
-      *flags |= SfMayBlock;
-}
-
-PRE(sys_msgrcv)
-{
-   PRINT("sys_msgrcv ( %ld, %#lx, %ld, %ld, %ld )",ARG1,ARG2,ARG3,ARG4,ARG5);
-   PRE_REG_READ5(long, "msgrcv",
-                 int, msqid, struct msgbuf *, msgp, vki_size_t, msgsz,
-                 long, msgytp, int, msgflg);
-   ML_(linux_PRE_sys_msgrcv)(tid, ARG1,ARG2,ARG3,ARG4,ARG5);
-   if ((ARG4 & VKI_IPC_NOWAIT) == 0)
-      *flags |= SfMayBlock;
-}
-POST(sys_msgrcv)
-{
-   ML_(linux_POST_sys_msgrcv)(tid, RES,ARG1,ARG2,ARG3,ARG4,ARG5);
-}
-
-PRE(sys_msgctl)
-{
-   PRINT("sys_msgctl ( %ld, %ld, %#lx )",ARG1,ARG2,ARG3);
-   PRE_REG_READ3(long, "msgctl",
-                 int, msqid, int, cmd, struct msqid_ds *, buf);
-   ML_(linux_PRE_sys_msgctl)(tid, ARG1,ARG2,ARG3);
-}
-POST(sys_msgctl)
-{
-   ML_(linux_POST_sys_msgctl)(tid, RES,ARG1,ARG2,ARG3);
-}
-
-PRE(sys_shmget)
-{
-   PRINT("sys_shmget ( %ld, %ld, %ld )",ARG1,ARG2,ARG3);
-   PRE_REG_READ3(long, "shmget", vki_key_t, key, vki_size_t, size, int, shmflg);
-}
-
-PRE(wrap_sys_shmat)
-{
-   UWord arg2tmp;
-   PRINT("wrap_sys_shmat ( %ld, %#lx, %ld )",ARG1,ARG2,ARG3);
-   PRE_REG_READ3(long, "shmat",
-                 int, shmid, const void *, shmaddr, int, shmflg);
-   arg2tmp = ML_(generic_PRE_sys_shmat)(tid, ARG1,ARG2,ARG3);
-   if (arg2tmp == 0)
-      SET_STATUS_Failure( VKI_EINVAL );
-   else
-      ARG2 = arg2tmp;  // used in POST
-}
-POST(wrap_sys_shmat)
-{
-   ML_(generic_POST_sys_shmat)(tid, RES,ARG1,ARG2,ARG3);
-}
-
-PRE(sys_shmdt)
-{
-   PRINT("sys_shmdt ( %#lx )",ARG1);
-   PRE_REG_READ1(long, "shmdt", const void *, shmaddr);
-   if (!ML_(generic_PRE_sys_shmdt)(tid, ARG1))
-      SET_STATUS_Failure( VKI_EINVAL );
-}
-POST(sys_shmdt)
-{
-   ML_(generic_POST_sys_shmdt)(tid, RES,ARG1);
-}
-
-PRE(sys_shmctl)
-{
-   PRINT("sys_shmctl ( %ld, %ld, %#lx )",ARG1,ARG2,ARG3);
-   PRE_REG_READ3(long, "shmctl",
-                 int, shmid, int, cmd, struct shmid_ds *, buf);
-   ML_(generic_PRE_sys_shmctl)(tid, ARG1,ARG2|VKI_IPC_64,ARG3);
-}
-POST(sys_shmctl)
-{
-   ML_(generic_POST_sys_shmctl)(tid, RES,ARG1,ARG2|VKI_IPC_64,ARG3);
-}
-
 PRE(sys_fadvise64)
 {
    PRINT("sys_fadvise64 ( %ld, %ld, %lu, %ld )", ARG1,ARG2,ARG3,ARG4);
@@ -1084,12 +740,12 @@ static SyscallTableEntry syscall_table[] = {
 
    GENX_(__NR_mremap,            sys_mremap),         // 25 
    GENX_(__NR_msync,             sys_msync),          // 26 
-   GENX_(__NR_mincore,           sys_mincore),        // 27 
+   GENXY(__NR_mincore,           sys_mincore),        // 27 
    GENX_(__NR_madvise,           sys_madvise),        // 28 
-   PLAX_(__NR_shmget,            sys_shmget),         // 29 
+   LINX_(__NR_shmget,            sys_shmget),         // 29 
 
-   PLAXY(__NR_shmat,             wrap_sys_shmat),     // 30 
-   PLAXY(__NR_shmctl,            sys_shmctl),         // 31 
+   LINXY(__NR_shmat,             wrap_sys_shmat),     // 30 
+   LINXY(__NR_shmctl,            sys_shmctl),         // 31 
    GENXY(__NR_dup,               sys_dup),            // 32 
    GENXY(__NR_dup2,              sys_dup2),           // 33 
    GENX_(__NR_pause,             sys_pause),          // 34 
@@ -1101,24 +757,24 @@ static SyscallTableEntry syscall_table[] = {
    GENX_(__NR_getpid,            sys_getpid),         // 39 
 
    LINXY(__NR_sendfile,          sys_sendfile),       // 40 
-   PLAXY(__NR_socket,            sys_socket),         // 41 
-   PLAX_(__NR_connect,           sys_connect),        // 42
-   PLAXY(__NR_accept,            sys_accept),         // 43 
-   PLAX_(__NR_sendto,            sys_sendto),         // 44 
+   LINXY(__NR_socket,            sys_socket),         // 41 
+   LINX_(__NR_connect,           sys_connect),        // 42
+   LINXY(__NR_accept,            sys_accept),         // 43 
+   LINX_(__NR_sendto,            sys_sendto),         // 44 
 
-   PLAXY(__NR_recvfrom,          sys_recvfrom),       // 45 
-   PLAX_(__NR_sendmsg,           sys_sendmsg),        // 46 
-   PLAXY(__NR_recvmsg,           sys_recvmsg),        // 47
-   PLAX_(__NR_shutdown,          sys_shutdown),       // 48 
-   PLAX_(__NR_bind,              sys_bind),           // 49 
+   LINXY(__NR_recvfrom,          sys_recvfrom),       // 45 
+   LINX_(__NR_sendmsg,           sys_sendmsg),        // 46 
+   LINXY(__NR_recvmsg,           sys_recvmsg),        // 47
+   LINX_(__NR_shutdown,          sys_shutdown),       // 48 
+   LINX_(__NR_bind,              sys_bind),           // 49 
 
-   PLAX_(__NR_listen,            sys_listen),         // 50 
-   PLAXY(__NR_getsockname,       sys_getsockname),    // 51 
-   PLAXY(__NR_getpeername,       sys_getpeername),    // 52 
-   PLAXY(__NR_socketpair,        sys_socketpair),     // 53 
-   PLAX_(__NR_setsockopt,        sys_setsockopt),     // 54
+   LINX_(__NR_listen,            sys_listen),         // 50 
+   LINXY(__NR_getsockname,       sys_getsockname),    // 51 
+   LINXY(__NR_getpeername,       sys_getpeername),    // 52 
+   LINXY(__NR_socketpair,        sys_socketpair),     // 53 
+   LINX_(__NR_setsockopt,        sys_setsockopt),     // 54
 
-   PLAXY(__NR_getsockopt,        sys_getsockopt),     // 55 
+   LINXY(__NR_getsockopt,        sys_getsockopt),     // 55 
    PLAX_(__NR_clone,             sys_clone),          // 56 
    GENX_(__NR_fork,              sys_fork),           // 57 
    GENX_(__NR_vfork,             sys_fork),           // 58 treat as fork
@@ -1128,16 +784,16 @@ static SyscallTableEntry syscall_table[] = {
    GENXY(__NR_wait4,             sys_wait4),          // 61 
    GENX_(__NR_kill,              sys_kill),           // 62 
    GENXY(__NR_uname,             sys_newuname),       // 63 
-   PLAX_(__NR_semget,            sys_semget),         // 64 
+   LINX_(__NR_semget,            sys_semget),         // 64 
 
-   PLAX_(__NR_semop,             sys_semop),          // 65 
-   PLAXY(__NR_semctl,            sys_semctl),         // 66 
-   PLAXY(__NR_shmdt,             sys_shmdt),          // 67 
-   PLAX_(__NR_msgget,            sys_msgget),         // 68 
-   PLAX_(__NR_msgsnd,            sys_msgsnd),         // 69 
+   LINX_(__NR_semop,             sys_semop),          // 65 
+   LINXY(__NR_semctl,            sys_semctl),         // 66 
+   LINXY(__NR_shmdt,             sys_shmdt),          // 67 
+   LINX_(__NR_msgget,            sys_msgget),         // 68 
+   LINX_(__NR_msgsnd,            sys_msgsnd),         // 69 
 
-   PLAXY(__NR_msgrcv,            sys_msgrcv),         // 70 
-   PLAXY(__NR_msgctl,            sys_msgctl),         // 71 
+   LINXY(__NR_msgrcv,            sys_msgrcv),         // 70 
+   LINXY(__NR_msgctl,            sys_msgctl),         // 71 
    LINXY(__NR_fcntl,             sys_fcntl),          // 72 
    GENX_(__NR_flock,             sys_flock),          // 73 
    GENX_(__NR_fsync,             sys_fsync),          // 74 
@@ -1256,7 +912,7 @@ static SyscallTableEntry syscall_table[] = {
    //   (__NR_swapoff,           sys_swapoff),        // 168 
    //   (__NR_reboot,            sys_reboot),         // 169 
 
-   //   (__NR_sethostname,       sys_sethostname),    // 170 
+   GENX_(__NR_sethostname,       sys_sethostname),    // 170 
    //   (__NR_setdomainname,     sys_setdomainname),  // 171 
    GENX_(__NR_iopl,              sys_iopl),           // 172 
    LINX_(__NR_ioperm,            sys_ioperm),         // 173 
@@ -1316,7 +972,7 @@ static SyscallTableEntry syscall_table[] = {
    LINX_(__NR_set_tid_address,   sys_set_tid_address),// 218 
    //   (__NR_restart_syscall,   sys_restart_syscall),// 219 
 
-   PLAX_(__NR_semtimedop,        sys_semtimedop),     // 220 
+   LINX_(__NR_semtimedop,        sys_semtimedop),     // 220 
    PLAX_(__NR_fadvise64,         sys_fadvise64),      // 221 
    LINXY(__NR_timer_create,      sys_timer_create),   // 222 
    LINXY(__NR_timer_settime,     sys_timer_settime),  // 223 
@@ -1383,24 +1039,24 @@ static SyscallTableEntry syscall_table[] = {
    LINXY(__NR_get_robust_list,	 sys_get_robust_list),  // 274
 
    LINX_(__NR_splice,            sys_splice),           // 275
-//   LINX_(__NR_tee,               sys_ni_syscall),       // 276
+   LINX_(__NR_tee,               sys_tee),              // 276
    LINX_(__NR_sync_file_range,   sys_sync_file_range),  // 277
-//   LINX_(__NR_vmsplice,          sys_ni_syscall),       // 278
-//   LINX_(__NR_move_pages,        sys_ni_syscall),       // 279
+   LINXY(__NR_vmsplice,          sys_vmsplice),         // 278
+   LINXY(__NR_move_pages,        sys_move_pages),       // 279
 
    LINX_(__NR_utimensat,         sys_utimensat),        // 280
    LINXY(__NR_epoll_pwait,       sys_epoll_pwait),      // 281
    LINXY(__NR_signalfd,          sys_signalfd),         // 282
    LINXY(__NR_timerfd_create,    sys_timerfd_create),   // 283
-   LINX_(__NR_eventfd,           sys_eventfd),          // 284
+   LINXY(__NR_eventfd,           sys_eventfd),          // 284
 
    LINX_(__NR_fallocate,         sys_fallocate),        // 285
    LINXY(__NR_timerfd_settime,   sys_timerfd_settime),  // 286
    LINXY(__NR_timerfd_gettime,   sys_timerfd_gettime),  // 287
-   PLAXY(__NR_accept4,           sys_accept4),          // 288
+   LINXY(__NR_accept4,           sys_accept4),          // 288
    LINXY(__NR_signalfd4,         sys_signalfd4),        // 289
 
-   LINX_(__NR_eventfd2,          sys_eventfd2),         // 290
+   LINXY(__NR_eventfd2,          sys_eventfd2),         // 290
    LINXY(__NR_epoll_create1,     sys_epoll_create1),    // 291
    LINXY(__NR_dup3,              sys_dup3),             // 292
    LINXY(__NR_pipe2,             sys_pipe2),            // 293
@@ -1409,7 +1065,23 @@ static SyscallTableEntry syscall_table[] = {
    LINXY(__NR_preadv,            sys_preadv),           // 295
    LINX_(__NR_pwritev,           sys_pwritev),          // 296
    LINXY(__NR_rt_tgsigqueueinfo, sys_rt_tgsigqueueinfo),// 297
-   LINXY(__NR_perf_counter_open, sys_perf_counter_open) // 298
+   LINXY(__NR_perf_event_open,   sys_perf_event_open),  // 298
+   LINXY(__NR_recvmmsg,          sys_recvmmsg),         // 299
+
+   LINXY(__NR_fanotify_init,     sys_fanotify_init),    // 300
+   LINX_(__NR_fanotify_mark,     sys_fanotify_mark),    // 301
+   LINXY(__NR_prlimit64,         sys_prlimit64),        // 302
+   LINXY(__NR_name_to_handle_at, sys_name_to_handle_at),// 303
+   LINXY(__NR_open_by_handle_at, sys_open_by_handle_at),// 304
+
+   LINXY(__NR_clock_adjtime,     sys_clock_adjtime),    // 305
+//   LINX_(__NR_syncfs,            sys_ni_syscall),       // 306
+   LINXY(__NR_sendmmsg,          sys_sendmmsg),         // 307
+//   LINX_(__NR_setns,             sys_ni_syscall),       // 308
+   LINXY(__NR_getcpu,            sys_getcpu),           // 309
+
+   LINXY(__NR_process_vm_readv,  sys_process_vm_readv), // 310
+   LINX_(__NR_process_vm_writev, sys_process_vm_writev) // 311
 };
 
 SyscallTableEntry* ML_(get_linux_syscall_entry) ( UInt sysno )
