@@ -8,7 +8,7 @@
    This file is part of Valgrind, a dynamic binary instrumentation
    framework.
 
-   Copyright (C) 2000-2007 Julian Seward
+   Copyright (C) 2000-2013 Julian Seward
       jseward@acm.org
 
    This program is free software; you can redistribute it and/or
@@ -51,15 +51,23 @@ typedef
 
 extern void VG_(load_suppressions)        ( void );
 
-extern void VG_(show_all_errors)          ( void );
+// if verbosity == 0,           print nothing.
+// else if xml                  print suppressions used (in xml format)
+// else if verbosity == 1       print Error summary
+// else                         print all errors and suppressions used.
+extern void VG_(show_all_errors)          ( Int verbosity, Bool xml );
+
+/* Print (in readable format) the last error that occured. */
+extern void VG_(show_last_error)          ( void );
 
 extern void VG_(show_error_counts_as_XML) ( void );
 
-extern Bool VG_(is_action_requested)      ( Char* action, Bool* clo );
+extern Bool VG_(is_action_requested)      ( const HChar* action, Bool* clo );
 
 extern Bool VG_(showing_core_errors)      ( void );
 
 extern UInt VG_(get_n_errs_found)         ( void );
+extern UInt VG_(get_n_errs_shown)         ( void );
 
 extern void VG_(print_errormgr_stats)     ( void );
 
