@@ -7,7 +7,7 @@
    This file is part of Valgrind, a dynamic binary instrumentation
    framework.
 
-   Copyright (C) 2000-2007 Julian Seward
+   Copyright (C) 2000-2013 Julian Seward
       jseward@acm.org
 
    This program is free software; you can redistribute it and/or
@@ -33,31 +33,47 @@
 
 //--------------------------------------------------------------------
 // PURPOSE: This module deals with:
+//
 // - code replacement: intercepting calls to client functions, and
 //   pointing them to a different piece of code.
+//
 // - loading notification: telling the core where certain client-space
 //   functions are when they get loaded.
+//
 // - function wrapping: add calls to code before and after client
 //   functions execute, for inspection and/or modification.
+//
+// - checking of --require-text-symbol= specifications: when a new
+//   object is loaded, its symbol table is examined, and if a symbol
+//   (as required by the specifications) is not found then the run
+//   is aborted.  See comment by VG_(clo_n_req_tsyms) in
+//   pub_core_options.h for background.  This doesn't have anything
+//   to do with function intercepting or wrapping, but it does have
+//   to do with examining all symbols at object load time, so this
+//   module seems like a logical place to put it.
+//
 //--------------------------------------------------------------------
 
 #include "pub_tool_redir.h"
-
+#include "pub_core_basics.h"      // Addr
+#include "pub_core_debuginfo.h"   // DebugInfo
 
 //--------------------------------------------------------------------
 // Notifications - by which we are told of state changes
 //--------------------------------------------------------------------
 
-/* Notify the module of a new SegInfo (called from m_debuginfo). */
-extern void VG_(redir_notify_new_SegInfo)( SegInfo* );
+/* Notify the module of a new DebugInfo (called from m_debuginfo). */
+extern void VG_(redir_notify_new_DebugInfo)( DebugInfo* );
 
-/* Notify the module of the disappearance of a SegInfo (also called
+/* Notify the module of the disappearance of a DebugInfo (also called
    from m_debuginfo). */
-extern void VG_(redir_notify_delete_SegInfo)( SegInfo* );
+extern void VG_(redir_notify_delete_DebugInfo)( DebugInfo* );
 
 /* Initialise the module, and load initial "hardwired" redirects. */
 extern void VG_(redir_initialise)( void );
 
+/* Notify the module of a new target for an indirect function. */
+extern void VG_(redir_add_ifunc_target)( Addr old_from, Addr new_from );
 
 //--------------------------------------------------------------------
 // Queries
@@ -113,7 +129,7 @@ extern Addr VG_(redir_do_lookup) ( Addr orig, Bool* isWrap );
 //extern Bool VG_(is_wrapper_return)(Addr eip);
 
 /* Primary interface for adding wrappers for client-side functions. */
-//extern CodeRedirect *VG_(add_wrapper)(const Char *from_lib, const Char *from_sym,
+//extern CodeRedirect *VG_(add_wrapper)(const HChar *from_lib, const HChar *from_sym,
 //				      const FuncWrapper *wrapper);
 //
 //extern Bool VG_(is_resolved)(const CodeRedirect *redir);
