@@ -7,7 +7,7 @@
    This file is part of Valgrind, a dynamic binary instrumentation
    framework.
 
-   Copyright (C) 2000-2008 Julian Seward
+   Copyright (C) 2000-2013 Julian Seward
       jseward@acm.org
 
    This program is free software; you can redistribute it and/or
@@ -40,6 +40,8 @@
 //--------------------------------------------------------------------
 
 #include "pub_tool_threadstate.h"
+#include "pub_core_libcsetjmp.h"   // VG_MINIMAL_JMP_BUF
+#include "pub_core_vki.h"          // vki_sigset_t
 
 /*------------------------------------------------------------*/
 /*--- Types                                                ---*/
@@ -68,7 +70,8 @@ typedef
    enum { 
       VgSrc_None,	 /* not exiting yet */
       VgSrc_ExitThread,  /* just this thread is exiting */
-      VgSrc_ExitProcess, /* entire process is exiting */
+      VgSrc_ExitProcess, /* this thread is exiting due to another thread
+                            calling exit() */
       VgSrc_FatalSig	 /* Killed by the default action of a fatal
 			    signal */
    }
@@ -83,10 +86,23 @@ typedef
    typedef VexGuestPPC32State VexGuestArchState;
 #elif defined(VGA_ppc64)
    typedef VexGuestPPC64State VexGuestArchState;
+#elif defined(VGA_arm)
+   typedef VexGuestARMState   VexGuestArchState;
+#elif defined(VGA_arm64)
+   typedef VexGuestARM64State VexGuestArchState;
+#elif defined(VGA_s390x)
+   typedef VexGuestS390XState VexGuestArchState;
+#elif defined(VGA_mips32)
+   typedef VexGuestMIPS32State VexGuestArchState;
+#elif defined(VGA_mips64)
+   typedef VexGuestMIPS64State VexGuestArchState;
 #else
 #  error Unknown architecture
 #endif
 
+/* Forward declarations */
+struct SyscallStatus;
+struct SyscallArgs;
 
 /* Architecture-specific thread state */
 typedef 
@@ -114,11 +130,13 @@ typedef
    ThreadArchState;
 
 
-/* OS-specific thread state */
+/* OS-specific thread state.  IMPORTANT: if you add fields to this,
+   you _must_ add code to os_state_clear() to initialise those
+   fields. */
 typedef
    struct {
       /* who we are */
-      Int lwpid;        // PID of kernel task
+      Int lwpid;        // PID of kernel task  (Darwin: Mach thread)
       Int threadgroup;  // thread group id
 
       ThreadId parent;  // parent tid (if any)
@@ -131,19 +149,134 @@ typedef
       Word exitcode; // in the case of exitgroup, set by someone else
       Int  fatalsig; // fatal signal
 
-#     if defined(VGO_aix5)
-      /* AIX specific fields to make thread cancellation sort-of work */
-      /* What is this thread's current cancellation state a la
-         POSIX (deferred vs async, enable vs disabled) ? */
-      Bool cancel_async;   // current cancel mode (async vs deferred)
-      Bool cancel_disabled; // cancellation disabled?
-      /* What's happened so far? */
-      enum { Canc_NoRequest=0, // no cancellation requested
-             Canc_Requested=1, // requested but not actioned
-             Canc_Actioned=2 } // requested and actioned
-           cancel_progress;
-      /* Initial state is False, False, Canc_Normal. */
+#     if defined(VGO_darwin)
+      // Mach trap POST handler as chosen by PRE
+      void (*post_mach_trap_fn)(ThreadId tid,
+                                struct SyscallArgs *, struct SyscallStatus *);
+    
+      // This thread's pthread
+      Addr pthread;
+    
+      // Argument passed when thread started
+      Addr func_arg;
+
+      // Synchronization between child thread and parent thread's POST wrapper
+      semaphore_t child_go;
+      semaphore_t child_done;
+
+      // Workqueue re-entry 
+      // (setjmp in PRE(workq_ops), longjmp in wqthread_hijack)
+      // DDD: JRS fixme: this comment is no longer correct; wq_jmpbuf is
+      // never used, and there is no such setjmp or longjmp pair.
+      // I guess we could leave wq_jmpbuf_valid in place though, since
+      // it does allow for an assertion in ML_(wqthread_continue_NORETURN).
+      Bool wq_jmpbuf_valid;
+      //jmp_buf wq_jmpbuf;
+
+      // Values saved from transient Mach RPC messages
+      Addr remote_port;  // destination for original message
+      Int msgh_id;       // outgoing message id
+      union {
+         struct {
+            Addr port;
+         } mach_port;
+         struct {
+            Int right;
+         } mach_port_allocate;
+         struct {
+            Addr port;
+            Int right;
+            Int delta;
+         } mach_port_mod_refs;
+         struct {
+            Addr task;
+            Addr name;
+            Int disposition;
+         } mach_port_insert_right;
+         struct {
+            Addr size;
+            int flags;
+         } vm_allocate;
+         struct {
+            Addr address;
+            Addr size;
+         } vm_deallocate;
+         struct {
+            Addr src;
+            Addr dst;
+            Addr size;
+         } vm_copy;
+         struct {
+            Addr address;
+            Addr size;
+            int set_maximum;
+            UWord new_protection;
+         } vm_protect;
+         struct {
+            Addr addr;
+            SizeT size;
+         } vm_read;
+         struct {
+            ULong addr;
+            ULong size;
+         } mach_vm_read;
+         struct {
+            Addr addr;
+            SizeT size;
+            Addr data;
+         } vm_read_overwrite;
+         struct {
+            Addr size;
+            int copy;
+            UWord protection;
+         } vm_map;
+         struct {
+            Addr size;
+         } vm_remap;
+         struct {
+            ULong size;
+            int flags;
+         } mach_vm_allocate;
+         struct {
+            ULong address;
+            ULong size;
+         } mach_vm_deallocate;
+         struct {
+            ULong address;
+            ULong size;
+            int set_maximum;
+            unsigned int new_protection;
+         } mach_vm_protect;
+         struct {
+            ULong size;
+            int copy;
+            UWord protection;
+         } mach_vm_map;
+         struct {
+            Addr thread;
+            UWord flavor;
+         } thread_get_state;
+         struct {
+            Addr address;
+         } io_connect_unmap_memory;
+         struct {
+            int which_port;
+         } task_get_special_port;
+         struct {
+            char *service_name;
+         } bootstrap_look_up;
+         struct {
+            vki_size_t size;
+         } WindowServer_29828;
+         struct {
+            Int access_rights;
+         } WindowServer_29831;
+         struct {
+            char *path;
+         } io_registry_entry_from_path;
+      } mach_args;
 #     endif
+
    }
    ThreadOSstate;
 
@@ -216,9 +349,20 @@ typedef struct {
    /* OS-specific thread state */
    ThreadOSstate os_state;
 
+   /* Error disablement level.  A counter which allows selectively
+      disabling error reporting in threads.  When zero, reporting is
+      enabled.  When nonzero, it is disabled.  This is controlled by
+      the client request 'VG_USERREQ__CHANGE_ERR_DISABLEMENT'.  New
+      threads are always created with this as zero (errors
+      enabled). */
+   UInt err_disablement_level;
+
    /* Per-thread jmp_buf to resume scheduler after a signal */
-   Bool    sched_jmpbuf_valid;
-   jmp_buf sched_jmpbuf;
+   Bool               sched_jmpbuf_valid;
+   VG_MINIMAL_JMP_BUF(sched_jmpbuf);
+
+   /* This thread's name. NULL, if no name. */
+   HChar *thread_name;
 }
 ThreadState;
 
@@ -236,13 +380,20 @@ extern ThreadState VG_(threads)[VG_N_THREADS];
 // to write to this.
 extern ThreadId VG_(running_tid);
 
+
 /*------------------------------------------------------------*/
 /*--- Basic operations on the thread table.                ---*/
 /*------------------------------------------------------------*/
 
+/* Initialize the m_threadstate module. */
+void VG_(init_Threads)(void);
+
 // Convert a ThreadStatus to a string.
 const HChar* VG_(name_of_ThreadStatus) ( ThreadStatus status );
 
+// Convert a VgSchedReturnCode to a string.
+const HChar* VG_(name_of_VgSchedReturnCode) ( VgSchedReturnCode retcode );
+
 /* Get the ThreadState for a particular thread */
 extern ThreadState *VG_(get_ThreadState) ( ThreadId tid );
 
