@@ -6,10 +6,10 @@
    This file is part of BBV, a Valgrind tool for generating SimPoint
    basic block vectors.
 
-   Copyright (C) 2006-2009 Vince Weaver
+   Copyright (C) 2006-2013 Vince Weaver
       vince _at_ csl.cornell.edu
 
-   pcfile code is Copyright (C) 2006-2009 Oriol Prat
+   pcfile code is Copyright (C) 2006-2013 Oriol Prat
       oriol.prat _at _ bsc.es
 
    This program is free software; you can redistribute it and/or
@@ -55,10 +55,10 @@
 static Int interval_size=DEFAULT_GRAIN_SIZE;
 
    /* filenames */
-static UChar *clo_bb_out_file="bb.out.%p";
-static UChar *clo_pc_out_file="pc.out.%p";
-static UChar *pc_out_file=NULL;
-static UChar *bb_out_file=NULL;
+static const HChar *clo_bb_out_file="bb.out.%p";
+static const HChar *clo_pc_out_file="pc.out.%p";
+static HChar *pc_out_file=NULL;
+static HChar *bb_out_file=NULL;
 
 
    /* output parameters */
@@ -66,7 +66,7 @@ static Bool instr_count_only=False;
 static Bool generate_pc_file=False;
 
    /* write buffer */
-static UChar buf[1024];
+static HChar buf[1024];
 
    /* Global values */
 static OSet* instr_info_table;  /* table that holds the basic block info */
@@ -95,7 +95,7 @@ struct BB_info {
    Int        block_num;         /* unique block identifier              */
    Int        *inst_counter;     /* times entered * num_instructions     */
    Bool       is_entry;          /* is this block a function entry point */
-   UChar      fn_name[FUNCTION_NAME_LENGTH];  /* Function block is in    */
+   HChar      fn_name[FUNCTION_NAME_LENGTH];  /* Function block is in    */
 };
 
 
@@ -137,7 +137,7 @@ static void dumpPcFile(void)
 static Int open_tracefile(Int thread_num)
 {
    SysRes  sres;
-   UChar temp_string[2048];
+   HChar temp_string[2048];
 
       /* For thread 1, don't append any thread number  */
       /* This lets the single-thread case not have any */
@@ -275,8 +275,8 @@ static Int get_inst_type(Int len, Addr addr)
 
 #if defined(VGA_x86) || defined(VGA_amd64)
 
-   unsigned char *inst_pointer;
-   unsigned char inst_byte;
+   UChar *inst_pointer;
+   UChar  inst_byte;
    int i,possible_rep;
 
    /* rep prefixed instructions are counted as one instruction on */
@@ -286,7 +286,7 @@ static Int get_inst_type(Int len, Addr addr)
    /*     SSE instructions.  So we need to specifically check for */
    /*     the following: movs, cmps, scas, lods, stos, ins, outs  */
 
-   inst_pointer=(unsigned char *)addr;
+   inst_pointer=(UChar *)addr;
    i=0;
    inst_byte=0;
    possible_rep=0;
@@ -322,7 +322,7 @@ static Int get_inst_type(Int len, Addr addr)
    /*     performance counters on pentium 4 processors so it is   */
    /*     useful to have that count when doing validation work.   */
 
-   inst_pointer=(unsigned char *)addr;
+   inst_pointer=(UChar *)addr;
    if (len>1) {
          /* FLDCW detection */
          /* opcode is 0xd9/5, ie 1101 1001 oo10 1mmm */
@@ -347,6 +347,7 @@ static Int get_inst_type(Int len, Addr addr)
 static IRSB* bbv_instrument ( VgCallbackClosure* closure,
                              IRSB* sbIn, VexGuestLayout* layout,
                              VexGuestExtents* vge,
+                             VexArchInfo* archinfo_host,
                              IRType gWordTy, IRType hWordTy )
 {
    Int      i,n_instrs=1;
@@ -530,7 +531,7 @@ static void bbv_post_clo_init(void)
 }
 
    /* Parse the command line options */
-static Bool bbv_process_cmd_line_option(Char* arg)
+static Bool bbv_process_cmd_line_option(const HChar* arg)
 {
    if VG_INT_CLO       (arg, "--interval-size",    interval_size) {}
    else if VG_STR_CLO  (arg, "--bb-out-file",      clo_bb_out_file) {}
@@ -607,7 +608,7 @@ static void bbv_pre_clo_init(void)
    VG_(details_version)         (NULL);
    VG_(details_description)     ("a SimPoint basic block vector generator");
    VG_(details_copyright_author)(
-      "Copyright (C) 2006-2009 Vince Weaver");
+      "Copyright (C) 2006-2013 Vince Weaver");
    VG_(details_bug_reports_to)  (VG_BUGS_TO);
 
    VG_(basic_tool_funcs)          (bbv_post_clo_init,
