@@ -7,7 +7,7 @@
    This file is part of Valgrind, a dynamic binary instrumentation
    framework.
 
-   Copyright (C) 2005-2010 Julian Seward
+   Copyright (C) 2005-2013 Julian Seward
       jseward@acm.org
 
    This program is free software; you can redistribute it and/or
@@ -64,8 +64,8 @@ typedef unsigned int vki_u32;
 
 /* PAGE_SHIFT determines the page size, unfortunately
    page size might vary between 32-bit and 64-bit ppc kernels */
-extern unsigned long VKI_PAGE_SHIFT;
-extern unsigned long VKI_PAGE_SIZE;
+extern UWord VKI_PAGE_SHIFT;
+extern UWord VKI_PAGE_SIZE;
 #define VKI_MAX_PAGE_SHIFT	16
 #define VKI_MAX_PAGE_SIZE	(1UL << VKI_MAX_PAGE_SHIFT)
 
@@ -360,6 +360,7 @@ struct vki_sigcontext {
 // From linux-2.6.13/include/asm-ppc64/fcntl.h
 //----------------------------------------------------------------------
 
+#define VKI_O_ACCMODE	         03
 #define VKI_O_RDONLY             00
 #define VKI_O_WRONLY             01
 #define VKI_O_RDWR               02
@@ -386,6 +387,18 @@ struct vki_sigcontext {
 #define VKI_F_SETSIG        10      /*  for sockets. */
 #define VKI_F_GETSIG        11      /*  for sockets. */
 
+#define VKI_F_SETOWN_EX		15
+#define VKI_F_GETOWN_EX		16
+
+#define VKI_F_OWNER_TID		0
+#define VKI_F_OWNER_PID		1
+#define VKI_F_OWNER_PGRP	2
+
+struct vki_f_owner_ex {
+	int	type;
+	__vki_kernel_pid_t	pid;
+};
+
 /* for F_[GET|SET]FL */
 #define VKI_FD_CLOEXEC  1  /* actually anything with low bit set goes */
 
@@ -410,6 +423,8 @@ struct vki_sigcontext {
 
 #define VKI_SO_TYPE         3
 
+#define VKI_SO_ATTACH_FILTER	26
+
 //----------------------------------------------------------------------
 // From linux-2.6.13/include/asm-ppc64/sockios.h
 //----------------------------------------------------------------------
@@ -589,6 +604,9 @@ struct vki_termios {
 // From linux-2.6.13/include/asm-ppc64/ioctls.h
 //----------------------------------------------------------------------
 
+#define VKI_FIOCLEX         _VKI_IO('f', 1)
+#define VKI_FIONCLEX        _VKI_IO('f', 2)
+
 #define VKI_TCGETS          _VKI_IOR('t', 19, struct vki_termios)
 #define VKI_TCSETS          _VKI_IOW('t', 20, struct vki_termios)
 #define VKI_TCSETSW         _VKI_IOW('t', 21, struct vki_termios)
@@ -620,6 +638,7 @@ struct vki_termios {
 #define VKI_FIOASYNC        _VKI_IOW('f', 125, int)
 #define VKI_TIOCSERGETLSR   0x5459 /* Get line status register */
 #define VKI_TIOCGICOUNT	    0x545D /* read serial port inline interrupt counts */
+#define VKI_FIOQSIZE        _VKI_IOR('f', 128, vki_loff_t)
 
 //----------------------------------------------------------------------
 // From linux-2.6.13/include/asm-ppc64/poll.h
@@ -760,6 +779,13 @@ struct vki_shminfo64 {
   unsigned long   __unused4;
 };
 
+//----------------------------------------------------------------------
+// From linux-2.6.8.1/include/asm-generic/errno.h
+//----------------------------------------------------------------------
+
+#define	VKI_ENOSYS       38  /* Function not implemented */
+#define	VKI_EOVERFLOW    75  /* Value too large for defined data type */
+
 //----------------------------------------------------------------------
 // end
 //----------------------------------------------------------------------
