@@ -7,7 +7,7 @@
    This file is part of Valgrind, a dynamic binary instrumentation
    framework.
 
-   Copyright (C) 2000-2007 Julian Seward 
+   Copyright (C) 2000-2013 Julian Seward 
       jseward@acm.org
 
    This program is free software; you can redistribute it and/or
@@ -63,6 +63,12 @@ typedef unsigned int vki_u32;
 #define VKI_MAX_PAGE_SHIFT	VKI_PAGE_SHIFT
 #define VKI_MAX_PAGE_SIZE	VKI_PAGE_SIZE
 
+//----------------------------------------------------------------------
+// From linux-2.6.35.4/arch/x86/include/asm/shmparam.h
+//----------------------------------------------------------------------
+
+#define VKI_SHMLBA  VKI_PAGE_SIZE
+
 //----------------------------------------------------------------------
 // From linux-2.6.8.1/include/asm-i386/signal.h
 //----------------------------------------------------------------------
@@ -149,6 +155,10 @@ typedef struct {
 #define VKI_SS_ONSTACK	1
 #define VKI_SS_DISABLE	2
 
+/* These are 'legacy' sigactions in which the size of sa_mask is fixed
+   (cannot be expanded at any future point) because it is sandwiched
+   between two other fields.
+   (there is identical kludgery in vki-ppc32-linux.h) */
 struct vki_old_sigaction {
         // [[Nb: a 'k' prefix is added to "sa_handler" because
         // bits/sigaction.h (which gets dragged in somehow via signal.h)
@@ -160,7 +170,7 @@ struct vki_old_sigaction {
         __vki_sigrestore_t sa_restorer;
 };
 
-struct vki_sigaction {
+struct vki_sigaction_base {
         // [[See comment about extra 'k' above]]
 	__vki_sighandler_t ksa_handler;
 	unsigned long sa_flags;
@@ -168,6 +178,12 @@ struct vki_sigaction {
 	vki_sigset_t sa_mask;		/* mask last for extensibility */
 };
 
+/* On Linux we use the same type for passing sigactions to
+   and from the kernel.  Hence: */
+typedef  struct vki_sigaction_base  vki_sigaction_toK_t;
+typedef  struct vki_sigaction_base  vki_sigaction_fromK_t;
+
+
 typedef struct vki_sigaltstack {
 	void __user *ss_sp;
 	int ss_flags;
@@ -262,6 +278,7 @@ struct vki_sigcontext {
 // From linux-2.6.8.1/include/asm-i386/fcntl.h
 //----------------------------------------------------------------------
 
+#define VKI_O_ACCMODE	     03
 #define VKI_O_RDONLY	     00
 #define VKI_O_WRONLY	     01
 #define VKI_O_RDWR	     02
@@ -292,6 +309,18 @@ struct vki_sigcontext {
 #define VKI_F_SETLK64		13
 #define VKI_F_SETLKW64		14
 
+#define VKI_F_SETOWN_EX		15
+#define VKI_F_GETOWN_EX		16
+
+#define VKI_F_OWNER_TID		0
+#define VKI_F_OWNER_PID		1
+#define VKI_F_OWNER_PGRP	2
+
+struct vki_f_owner_ex {
+	int	type;
+	__vki_kernel_pid_t	pid;
+};
+
 /* for F_[GET|SET]FL */
 #define VKI_FD_CLOEXEC	1	/* actually anything with low bit set goes */
 
@@ -314,13 +343,16 @@ struct vki_sigcontext {
 
 #define VKI_SO_TYPE	3
 
+#define VKI_SO_ATTACH_FILTER	26
+
 //----------------------------------------------------------------------
 // From linux-2.6.8.1/include/asm-i386/sockios.h
 //----------------------------------------------------------------------
 
-#define VKI_SIOCSPGRP	0x8902
-#define VKI_SIOCGPGRP	0x8904
-#define VKI_SIOCGSTAMP	0x8906		/* Get stamp */
+#define VKI_SIOCSPGRP		0x8902
+#define VKI_SIOCGPGRP		0x8904
+#define VKI_SIOCGSTAMP		0x8906		/* Get stamp (timeval) */
+#define VKI_SIOCGSTAMPNS	0x8907		/* Get stamp (timespec) */
 
 //----------------------------------------------------------------------
 // From linux-2.6.8.1/include/asm-i386/stat.h
@@ -468,13 +500,6 @@ struct vki_termios {
 	 ((nr)   << _VKI_IOC_NRSHIFT) | \
 	 ((size) << _VKI_IOC_SIZESHIFT))
 
-/* provoke compile error for invalid uses of size argument */
-extern unsigned int __vki_invalid_size_argument_for_IOC;
-#define _VKI_IOC_TYPECHECK(t) \
-	((sizeof(t) == sizeof(t[1]) && \
-	  sizeof(t) < (1 << _VKI_IOC_SIZEBITS)) ? \
-	  sizeof(t) : __vki_invalid_size_argument_for_IOC)
-
 /* used to create numbers */
 #define _VKI_IO(type,nr)	_VKI_IOC(_VKI_IOC_NONE,(type),(nr),0)
 #define _VKI_IOR(type,nr,size)	_VKI_IOC(_VKI_IOC_READ,(type),(nr),(_VKI_IOC_TYPECHECK(size)))
@@ -519,6 +544,8 @@ extern unsigned int __vki_invalid_size_argument_for_IOC;
 #define VKI_TIOCGPTN	_VKI_IOR('T',0x30, unsigned int) /* Get Pty Number (of pty-mux device) */
 #define VKI_TIOCSPTLCK	_VKI_IOW('T',0x31, int)  /* Lock/unlock Pty */
 
+#define VKI_FIONCLEX    0x5450
+#define VKI_FIOCLEX     0x5451
 #define VKI_FIOASYNC	0x5452
 #define VKI_TIOCSERGETLSR   0x5459 /* Get line status register */
 
@@ -789,6 +816,8 @@ struct vki_shminfo64 {
 #define VKI_PTRACE_SETFPREGS          15
 #define VKI_PTRACE_GETFPXREGS         18
 #define VKI_PTRACE_SETFPXREGS         19
+#define VKI_PTRACE_GET_THREAD_AREA    25
+#define VKI_PTRACE_SET_THREAD_AREA    26
 
 //----------------------------------------------------------------------
 // From linux-2.6.15.4/include/asm-i386/vm86.h
@@ -864,6 +893,13 @@ struct vki_vm86plus_struct {
 	struct vki_vm86plus_info_struct vm86plus;
 };
 
+//----------------------------------------------------------------------
+// From linux-2.6.8.1/include/asm-generic/errno.h
+//----------------------------------------------------------------------
+
+#define	VKI_ENOSYS       38  /* Function not implemented */
+#define	VKI_EOVERFLOW    75  /* Value too large for defined data type */
+
 //----------------------------------------------------------------------
 // And that's it!
 //----------------------------------------------------------------------
