@@ -264,6 +264,8 @@ test_stpcpy (void)
   SIMPLE_COPY(stpcpy, 16, "6666666666666666", 59);
 }
 
+// DDD: better done by testing for the function.
+#if !defined(__APPLE__)
 static void
 test_stpncpy (void)
 {
@@ -278,6 +280,7 @@ test_stpncpy (void)
   check (stpncpy (one, "abcd", 6) == one + 4, 7);
   check (one[4] == '\0' && one[5] == '\0' && one[6] == 'x', 8);
 }
+#endif
 
 static void
 test_strcat (void)
@@ -463,6 +466,8 @@ test_strchr (void)
    }
 }
 
+// DDD: better done by testing for the function.
+#if !defined(__APPLE__)
 static void
 test_strchrnul (void)
 {
@@ -496,7 +501,10 @@ test_strchrnul (void)
       }
    }
 }
+#endif
 
+// DDD: better done by testing for the function.
+#if !defined(__APPLE__)
 static void
 test_rawmemchr (void)
 {
@@ -523,6 +531,7 @@ test_rawmemchr (void)
       }
    }
 }
+#endif
 
 static void
 test_index (void)
@@ -570,6 +579,8 @@ test_strrchr (void)
    }
 }
 
+// DDD: better done by testing for the function.
+#if !defined(__APPLE__)
 static void
 test_memrchr (void)
 {
@@ -615,6 +626,7 @@ test_memrchr (void)
     }
   }
 }
+#endif
 
 static void
 test_rindex (void)
@@ -890,6 +902,7 @@ test_strsep (void)
   equal(one+4, "c", 50);
 
   {
+#   if !defined(__APPLE__)
     char text[] = "This,is,a,test";
     char *list = strdupa (text);
     equal (strsep (&list, ","), "This", 51);
@@ -897,6 +910,7 @@ test_strsep (void)
     equal (strsep (&list, ","), "a", 53);
     equal (strsep (&list, ","), "test", 54);
     check (strsep (&list, ",") == NULL, 55);
+#   endif
   }
 
   cp = strcpy(one, "a,b, c,, ,d,");
@@ -1049,6 +1063,7 @@ test_memcpy (void)
     }
 }
 
+#if !defined(__APPLE__)
 static void
 test_mempcpy (void)
 {
@@ -1085,6 +1100,7 @@ test_mempcpy (void)
       equal (two, "hi there", 12 + (i * 6));
     }
 }
+#endif
 
 static void
 test_memmove (void)
@@ -1263,6 +1279,7 @@ test_bzero (void)
   equal(one, "abcdef", 4);		/* Zero-length copy. */
 }
 
+#if !defined(__APPLE__)
 static void
 test_strndup (void)
 {
@@ -1286,6 +1303,7 @@ test_strndup (void)
     equal(p, "abc", 6);
   free (p);
 }
+#endif
 
 static void
 test_bcmp (void)
@@ -1355,6 +1373,33 @@ test_strncasecmp (void)
   check(strncasecmp("ADC", "abcd", 2) > 0, 20);
 }
 
+static void
+test_strcasestr (void)
+{
+  it = "strcasestr";
+  check(strcasestr("abCd", "z") == NULL, 1);	/* Not found. */
+  check(strcasestr("AbcD", "abX") == NULL, 2);	/* Dead end. */
+  (void) strcpy(one, "abCd");
+  check(strcasestr(one, "c") == one+2, 3);	/* Basic test. */
+  check(strcasestr(one, "Bc") == one+1, 4);	/* Multichar. */
+  check(strcasestr(one, "d") == one+3, 5);	/* End of string. */
+  check(strcasestr(one, "Cd") == one+2, 6);	/* Tail of string. */
+  check(strcasestr(one, "aBc") == one, 7);	/* Beginning. */
+  check(strcasestr(one, "aBcd") == one, 8);	/* Exact match. */
+  check(strcasestr(one, "AbcDe") == NULL, 9);	/* Too long. */
+  check(strcasestr(one, "dE") == NULL, 10);	/* Past end. */
+  check(strcasestr(one, "") == one, 11);	/* Finding empty. */
+  (void) strcpy(one, "abAba");
+  check(strcasestr(one, "Ba") == one+1, 12);	/* Finding first. */
+  (void) strcpy(one, "");
+  check(strcasestr(one, "b") == NULL, 13);	/* Empty string. */
+  check(strcasestr(one, "") == one, 14);	/* Empty in empty string. */
+  (void) strcpy(one, "BcbCa");
+  check(strcasestr(one, "bCa") == one+2, 15);	/* False start. */
+  (void) strcpy(one, "bbBcaBbcA");
+  check(strcasestr(one, "bbCa") == one+1, 16);	/* With overlap. */
+}
+
 int
 main (void)
 {
@@ -1369,8 +1414,10 @@ main (void)
   /* A closely related function is stpcpy.  */
   test_stpcpy ();
 
+#if !defined(__APPLE__)
   /* stpncpy.  */
   test_stpncpy ();
+#endif
 
   /* strcat.  */
   test_strcat ();
@@ -1390,11 +1437,15 @@ main (void)
   /* strchr.  */
   test_strchr ();
 
+# if !defined(__APPLE__)
   /* strchrnul.  */
   test_strchrnul ();
+# endif
 
+# if !defined(__APPLE__)
   /* rawmemchr.  */
   test_rawmemchr ();
+# endif
 
   /* index - just like strchr.  */
   test_index ();
@@ -1402,8 +1453,10 @@ main (void)
   /* strrchr.  */
   test_strrchr ();
 
+# if !defined(__APPLE__)
   /* memrchr.  */
   test_memrchr ();
+# endif
 
   /* rindex - just like strrchr.  */
   test_rindex ();
@@ -1441,8 +1494,10 @@ main (void)
   /* memmove - must work on overlap.  */
   test_memmove ();
 
+# if !defined(__APPLE__)
   /* mempcpy */
   test_mempcpy ();
+# endif
 
   /* memccpy.  */
   test_memccpy ();
@@ -1459,8 +1514,10 @@ main (void)
   /* bcmp - somewhat like memcmp.  */
   test_bcmp ();
 
+#if !defined(__APPLE__)
   /* strndup.  */
   test_strndup ();
+#endif
 
   /* strerror - VERY system-dependent.  */
   test_strerror ();
@@ -1471,6 +1528,8 @@ main (void)
   /* strncasecmp.  Without locale dependencies.  */
   test_strncasecmp ();
 
+  test_strcasestr ();
+
   if (errors == 0)
     {
       status = EXIT_SUCCESS;
@@ -1479,7 +1538,7 @@ main (void)
   else
     {
       status = EXIT_FAILURE;
-      printf("%Zd errors.\n", errors);
+      printf("%d errors.\n", (int)errors);
     }
 
   return status;
