@@ -6,7 +6,9 @@ int main(int argc, char **argv)
 {
    if (argc == 1)
    {
-      if (execve(argv[0], NULL, NULL) < 0)
+      // This tests the case where argv and envp are NULL, which is easy to
+      // get wrong because it's an unusual case.
+      if (execve("/bin/true", NULL, NULL) < 0)
       {
          perror("execve");
          exit(1);
